from __future__ import annotations

import getpass
import inspect
import typing
from loguru import logger

if typing.TYPE_CHECKING:
    from telethon import TelegramClient


class TelethonUtils:
    def __init__(self, client: TelegramClient):
        self.client = client

    def start(
        self,
        phone: typing.Union[typing.Callable[[], str], str] = lambda: input(
            "Please enter your phone (or bot token): "
        ),
        password: typing.Union[typing.Callable[[], str], str] = lambda: getpass.getpass(
            "Please enter your password: "
        ),
        *,
        bot_token: str = None,
        force_sms: bool = False,
        code_callback: typing.Callable[[], typing.Union[str, int]] = None,
        first_name: str = "New User",
        last_name: str = "",
        max_attempts: int = 3,
    ):
        """This function is a modified version of telethon.client.auth.AuthMethods.start"""

        if code_callback is None:

            def code_callback():
                return input("Please enter the code you received: ")

        elif not callable(code_callback):
            raise ValueError(
                "The code_callback parameter needs to be a callable "
                "function that returns the code you received by Telegram."
            )

        if not phone and not bot_token:
            raise ValueError("No phone number or bot token provided.")

        if phone and bot_token and not callable(phone):
            raise ValueError("Both a phone and a bot token provided, " "must only provide one of either")

        coro = self._start(
            phone=phone,
            password=password,
            bot_token=bot_token,
            force_sms=force_sms,
            code_callback=code_callback,
            first_name=first_name,
            last_name=last_name,
            max_attempts=max_attempts,
        )
        return coro if self.client.loop.is_running() else self.client.loop.run_until_complete(coro)

    async def _start(
        self,
        phone,
        password,
        bot_token,
        force_sms,
        code_callback,
        first_name,
        last_name,
        max_attempts,
    ):
        """This function is a modified version of telethon.client.auth.AuthMethods._start"""

        from telethon import errors, utils

        if not self.client.is_connected():
            await self.client.connect()

        # Rather than using `is_user_authorized`, use `get_me`. While this is
        # more expensive and needs to retrieve more data from the server, it
        # enables the library to warn users trying to login to a different
        # account. See #1172.
        me = await self.client.get_me()
        if me is not None:
            # The warnings here are on a best-effort and may fail.
            if bot_token:
                # bot_token's first part has the bot ID, but it may be invalid
                # so don't try to parse as int (instead cast our ID to string).
                if bot_token[: bot_token.find(":")] != str(me.id):
                    logger.warning(
                        "该会话已经包含一个授权用户,所以没有使用提供的 bot_token 登录机器人账号(可能导致没有使用你期望的用户)"
                    )
            elif phone and not callable(phone) and utils.parse_phone(phone) != me.phone:
                logger.warning(
                    "该会话已经包含一个授权用户,所以没有使用提供的手机号登录账号 (可能导致没有使用你期望的用户)"
                )

            return self.client

        if not bot_token:
            # Turn the callable into a valid phone number (or bot token)
            while callable(phone):
                value = phone()
                if inspect.isawaitable(value):
                    value = await value

                if ":" in value:
                    # Bot tokens have 'user_id:access_hash' format
                    bot_token = value
                    break

                phone = utils.parse_phone(value) or phone

        if bot_token:
            await self.client.sign_in(bot_token=bot_token)
            return self.client

        me = None
        attempts = 0
        two_step_detected = False

        await self.client.send_code_request(phone, force_sms=force_sms)
        while attempts < max_attempts:
            try:
                value = code_callback()
                if inspect.isawaitable(value):
                    value = await value

                # Since sign-in with no code works (it sends the code)
                # we must double-check that here. Else we'll assume we
                # logged in, and it will return None as the User.
                if not value:
                    raise errors.PhoneCodeEmptyError(request=None)

                # Raises SessionPasswordNeededError if 2FA enabled
                me = await self.client.sign_in(phone, code=value)
                break
            except errors.SessionPasswordNeededError:
                two_step_detected = True
                break
            except (
                errors.PhoneCodeEmptyError,
                errors.PhoneCodeExpiredError,
                errors.PhoneCodeHashEmptyError,
                errors.PhoneCodeInvalidError,
            ):
                logger.warning("验证码错误, 请重试.")

            attempts += 1
        else:
            raise RuntimeError("{} consecutive sign-in attempts failed. Aborting".format(max_attempts))

        if two_step_detected:
            if not password:
                raise ValueError(
                    "Two-step verification is enabled for this account. "
                    "Please provide the 'password' argument to 'start()'."
                )

            if callable(password):
                for _ in range(max_attempts):
                    try:
                        value = password()
                        if inspect.isawaitable(value):
                            value = await value

                        me = await self.client.sign_in(phone=phone, password=value)
                        break
                    except errors.PasswordHashInvalidError:
                        logger.warning("两步验证密码错误, 请重试.")
                else:
                    raise errors.PasswordHashInvalidError(request=None)
            else:
                me = await self.client.sign_in(phone=phone, password=password)

        return self.client



from . import BotCheckin

__ignore__ = True


class OixelCheckin(BotCheckin):
    name = "Oixel"
    bot_username = "oixel_bot"
    bot_checkin_cmd = "📅 签到"
    additional_auth = ["prime"]
    max_retries = 6



from enum import IntEnum
from typing import Optional, Union
from io import BytesIO
from multiprocessing import Process, Queue
import asyncio
import time
import uuid

from .data import get_datas


class CharRange(IntEnum):
    NUMBER = 0
    LLETTER = 1
    ULETTER = 2
    LLETTER_ULETTER = 3
    NUMBER_LLETTER = 4
    NUMBER_ULETTER = 5
    NUMBER_LLETTER_ULETTER = 6
    NOT_NUMBER_LLETTER_ULETTER = 7


class OCRService:
    _pool = {}
    _pool_lock = asyncio.Lock()

    @classmethod
    async def get(
        cls,
        ocr_name: str = None,
        char_range: Optional[Union[CharRange, str]] = None,
    ):
        # 创建用于标识唯一实例的键
        key = (ocr_name, char_range)
        async with cls._pool_lock:
            # 检查池中是否存在相同配置的实例
            if key in cls._pool:
                return cls._pool[key]
            instance = cls(ocr_name, char_range)
            cls._pool[key] = instance
            return instance

    def __init__(
        self,
        ocr_name: str = None,
        char_range: Optional[Union[CharRange, str]] = None,
    ) -> None:
        self.ocr_name = ocr_name
        self.char_range = char_range

        self._process = None
        self._queue_in = None  # 发送图片数据的队列
        self._queue_out = None  # 接收识别结果的队列
        self._subscribers = 0
        self._last_active = time.time()
        self._stop_event = None
        self._monitor_task = None
        self._pending_requests = {}  # 存储待处理的请求

    async def start(self):
        """启动OCR进程"""
        if self._process and self._process.is_alive():
            return

        self._queue_in = Queue()
        self._queue_out = Queue()
        self._stop_event = asyncio.Event()

        self._process = Process(
            target=self._process_main,
            args=(
                self._queue_in,
                self._queue_out,
                self.ocr_name,
                self.char_range,
            ),
            daemon=True,
        )
        self._process.start()

        # 启动监控任务
        self._monitor_task = asyncio.create_task(self._monitor())

    async def stop(self, force: bool = False):
        """停止OCR进程"""
        if force:
            await self.force_stop()
        else:
            self._subscribers = 0  # 这将触发监控任务在空闲超时后关闭进程

    async def force_stop(self):
        """强制停止OCR进程"""
        if self._monitor_task:
            self._monitor_task.cancel()
            try:
                await self._monitor_task
            except asyncio.CancelledError:
                pass

        # 处理所有未完成的请求
        for future in self._pending_requests.values():
            if not future.done():
                future.set_exception(Exception("OCR进程已停止"))
        self._pending_requests.clear()

        if self._process and self._process.is_alive():
            self._queue_in.put(("stop", None))
            self._process.join(timeout=1)
            if self._process.is_alive():
                self._process.terminate()
                self._process.join()

        self._process = None
        self._queue_in = None
        self._queue_out = None
        self._stop_event = None

    async def run(self, image_data: BytesIO, timeout: int = 60, gif: bool = False) -> str:
        """发送图片到OCR进程并等待结果"""
        if not self._process or not self._process.is_alive():
            await self.start()

        # 生成唯一请求ID
        request_id = str(uuid.uuid4())
        future = asyncio.Future()
        self._pending_requests[request_id] = future

        try:
            self._last_active = time.time()
            self._queue_in.put(("process", (request_id, image_data.getvalue(), gif)))
            result = await asyncio.wait_for(future, timeout=timeout)
            return result
        finally:
            self._pending_requests.pop(request_id, None)

    def subscribe(self):
        """增加使用者计数"""
        self._subscribers += 1
        self._last_active = time.time()

    def unsubscribe(self):
        """减少使用者计数"""
        self._subscribers = max(0, self._subscribers - 1)
        self._last_active = time.time()

    async def _monitor(self):
        """监控进程状态和空闲时间，同时处理返回结果"""
        while True:
            try:
                # 检查进程状态和空闲超时
                if self._subscribers == 0 and time.time() - self._last_active > 300:
                    await self.force_stop()
                    break

                if self._process and not self._process.is_alive():
                    await self.force_stop()
                    break

                # 非阻塞方式检查结果队列
                try:
                    status, (request_id, result) = await asyncio.get_event_loop().run_in_executor(
                        None, self._queue_out.get_nowait
                    )

                    # 找到对应的future并设置结果
                    if request_id in self._pending_requests:
                        future = self._pending_requests[request_id]
                        if status == "error":
                            future.set_exception(Exception(result))
                        else:
                            future.set_result(result)
                except (asyncio.CancelledError, Exception):
                    pass

                await asyncio.sleep(0.1)  # 短暂休眠避免CPU占用过高

            except asyncio.CancelledError:
                break

    @staticmethod
    def _process_main(*args, **kw):
        return asyncio.run(OCRService._async_process_main(*args, **kw))

    @staticmethod
    async def _async_process_main(
        queue_in: Queue,
        queue_out: Queue,
        ocr_name: str,
        char_range: Optional[Union[CharRange, str]],
    ):
        model = None
        use_probability = False

        try:
            from ddddocr import DdddOcr
            from onnxruntime.capi.onnxruntime_pybind11_state import InvalidProtobuf
            from PIL import Image
            from io import BytesIO

            # 加载模型
            if not ocr_name:
                model = DdddOcr(beta=True, show_ad=False)
                if char_range:
                    use_probability = True
                    model.set_ranges(char_range)
            else:
                data = []
                files = (f"{ocr_name}.onnx", f"{ocr_name}.json")
                async for p in get_datas(files, caller="OCR"):
                    if p is None:
                        queue_out.put(("error", "无法下载所需文件"))
                        return
                    data.append(p)
                try:
                    model = DdddOcr(show_ad=False, import_onnx_path=str(data[0]), charsets_path=str(data[1]))
                except InvalidProtobuf:
                    queue_out.put(("error", "文件下载不完全"))
                    return

            def process_single_image(image, use_probability):
                if use_probability:
                    ocr_result = model.classification(image, probability=True)
                    ocr_text = ""
                    for i in ocr_result["probability"]:
                        ocr_text += ocr_result["charsets"][i.index(max(i))]
                else:
                    ocr_text = model.classification(image)
                return ocr_text

            def process_gif(gif_data):
                gif = Image.open(BytesIO(gif_data))
                frame_count = gif.n_frames

                # Calculate how many frames to use (up to 5)
                num_frames = min(5, frame_count)
                frame_indices = [i * (frame_count - 1) // (num_frames - 1) for i in range(num_frames)]

                # Get the first frame to determine size
                gif.seek(0)
                base_frame = gif.copy().convert("RGBA")

                # Create a blank transparent image
                composite = Image.new("RGBA", base_frame.size, (0, 0, 0, 0))

                # Calculate alpha for each frame
                alpha_per_frame = 255 // num_frames

                for idx in frame_indices:
                    gif.seek(idx)
                    frame = gif.copy().convert("RGBA")
                    # Apply partial transparency
                    frame.putalpha(alpha_per_frame)
                    composite = Image.alpha_composite(composite, frame)

                # Convert final image to RGB for OCR
                final_image = composite.convert("RGB")
                final_bytes = BytesIO()
                final_image.save(final_bytes, format="PNG")
                return process_single_image(final_bytes.getvalue(), use_probability)

            # 处理请求循环
            while True:
                try:
                    cmd, data = queue_in.get()
                except KeyboardInterrupt:
                    break
                if cmd == "stop":
                    break

                request_id, image_data, is_gif = data
                try:
                    if is_gif:
                        ocr_text = process_gif(image_data)
                    else:
                        image = Image.open(BytesIO(image_data))
                        ocr_text = process_single_image(image, use_probability)
                    queue_out.put(("success", (request_id, ocr_text)))
                except Exception as e:
                    queue_out.put(("error", (request_id, str(e))))

        finally:
            if model:
                del model

    def __enter__(self):
        """上下文管理器入口"""
        self.subscribe()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """上下文管理器出口"""
        self.unsubscribe()
        return False  # 返回False允许异常正常传播



import importlib as _i
import typing as _t

if _t.TYPE_CHECKING:
    from ._base import BaseBotCheckin, BotCheckin, AnswerBotCheckin, MessageType
    from ._templ_a import TemplateACheckin
    from ._templ_b import TemplateBCheckin

MENU = {
    "._base": ["BaseBotCheckin", "BotCheckin", "AnswerBotCheckin", "MessageType"],
    "._templ_a": ["TemplateACheckin"],
    "._templ_b": ["TemplateBCheckin"],
}


def __getattr__(spec: str):
    for module, specs in MENU.items():
        if isinstance(specs, dict):
            for s, t in specs.items():
                if s == spec:
                    m = _i.import_module(module, package=__name__)
                    return getattr(m, t or s)
        else:
            if spec in specs:
                m = _i.import_module(module, package=__name__)
                return getattr(m, spec)
    else:
        try:
            m = _i.import_module(f".{spec}", package=__name__)
            return m
        except ImportError:
            raise AttributeError(f"module '{__name__}' has no attribute '{spec}'") from None



import asyncio
from datetime import datetime
import random
import re

from pyrogram.types import Message
from pyrogram.errors import RPCError, MessageIdInvalid

from embykeeper.utils import to_iterable, truncate_str
from embykeeper.cache import cache

from ..link import Link
from ..lock import pornemby_alert
from . import Monitor


QA_CACHE_KEY = "monitor.pornemby.answer.qa"


class _PornembyAnswerResultMonitor(Monitor):
    name = "Pornemby 问题答案"
    chat_except_keyword = "猜猜是什么番号"
    chat_keyword = r"问题\d*：(.*?)\n+A:(.*)\n+B:(.*)\n+C:(.*)\n+D:(.*)\n+答案为：([ABCD])"
    additional_auth = ["pornemby_pack"]
    allow_edit = True
    allow_caption = False

    key_map = {"A": 1, "B": 2, "C": 3, "D": 4}

    async def on_trigger(self, message: Message, key, reply):
        spec = f"[gray50]({truncate_str(key[0], 10)})[/]"
        self.log.info(f"本题正确答案为 {key[5]} ({key[self.key_map[key[5]]]}): {spec}.")


class _PornembyAnswerAnswerMonitor(Monitor):
    name = "Pornemby 问题回答"
    history_chat_name = "Pornemby"
    chat_user = [
        "pornemby_question_bot",
        "PronembyTGBot2_bot",
        "PronembyTGBot3_bot",
        "PornembyBot",
        "Porn_Emby_Bot",
        "Porn_Emby_Scriptbot",
    ]
    chat_except_keyword = "猜猜是什么番号"
    chat_keyword = r"问题\d*：(.*?)(\(.*第\d+题.*\))\n+(A:.*\n+B:.*\n+C:.*\n+D:.*)\n(?!\n*答案)"
    additional_auth = ["pornemby_pack"]

    lock = asyncio.Lock()

    key_map = {
        "A": ["A", "🅰"],
        "B": ["B", "🅱"],
        "C": ["C", "🅲"],
        "D": ["D", "🅳"],
    }

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.update_task = None

    async def update_cache(self, to_date=None):
        if not to_date:
            to_date = datetime.fromtimestamp(cache.get(f"{QA_CACHE_KEY}.timestamp", 0))

        if not to_date:
            self.log.info("首次使用 Pornemby 科举, 正在缓存问题答案历史.")
        else:
            self.log.info(f"正在更新问题答案历史缓存.")
            self.log.debug(f"上一次问题答案历史写入于 {to_date.strftime('%Y-%m-%d %H:%M')}.")

        count = 0
        qs = 0
        finished = False
        while not finished:
            finished = True
            m: Message
            for g in to_iterable(self.history_chat_name):
                async for m in self.client.search_messages(g, limit=100, offset=count, query="答案为"):
                    if m.date < to_date:
                        break
                    count += 1
                    finished = False
                    if m.text:
                        for key in _PornembyAnswerResultMonitor.keys(m):
                            qs += 1
                            cache.set(f"{QA_CACHE_KEY}.data.{key[0]}", key[5])
            if count and (finished or count % 500 == 0):
                self.log.info(f"读取问题答案历史: 已读取 {qs} 问题 / {count} 信息.")
                await asyncio.sleep(2)
        self.log.debug(f"已向问题答案历史缓存写入 {qs} 条问题.")
        cache.set(f"{QA_CACHE_KEY}.timestamp", datetime.now().timestamp())

    async def update(self):
        try:
            await asyncio.wait_for(self.lock.acquire(), 1)
        except asyncio.TimeoutError:
            self.log.debug("等待其他协程缓存问题答案历史.")
            async with self.lock:
                return True
        else:
            try:
                await self.update_cache()
                return True
            finally:
                self.lock.release()

    async def cache_watchdog(self):
        while True:
            secs = 3600 * 12
            self.log.debug(f"等待 {secs} 秒后进行缓存更新.")
            await asyncio.sleep(secs)
            await self.update()

    async def init(self):
        self.update_task = asyncio.create_task(self.cache_watchdog())
        return await self.update()

    async def on_trigger(self, message: Message, key, reply):
        spec = f"[gray50]({truncate_str(key[0], 10)})[/]"
        if pornemby_alert.get(self.client.me.id, False):
            self.log.info(f"由于风险急停不作答: {spec}.")
            return
        if random.random() > self.config.get("possibility", 1.0):
            self.log.info(f"由于概率设置不作答: {spec}.")
            return
        result = cache.get(f"{QA_CACHE_KEY}.data.{key[0]}")
        if result:
            self.log.info(f"从缓存回答问题为{result}: {spec}.")
        elif self.config.get("only_history", False):
            self.log.info(f"未从历史缓存找到问题, 请自行回答: {spec}.")
            return
        else:
            question = key[0]
            choices = key[2]
            question = re.sub(r"\([^\)]*From资料库:第\d+题\)", "", question)
            for _ in range(3):
                self.log.debug(f"未从历史缓存找到问题, 开始请求云端问题回答: {spec}.")
                result, by = await Link(self.client).pornemby_answer(question + "\n" + choices)
                if result:
                    self.log.info(f"请求 {by or '云端'} 问题回答为 {result}: {spec}.")
                    break
                else:
                    self.log.info(f"云端问题回答错误或超时, 正在重试: {spec}.")
            else:
                self.log.info(f"错误次数超限, 回答失败: {spec}.")
                return
        try:
            await asyncio.sleep(random.uniform(2, 4))
            buttons = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            answer_options = self.key_map[result]
            for button_text in buttons:
                if any((o in button_text) for o in answer_options):
                    try:
                        await message.click(button_text)
                    except (TimeoutError, MessageIdInvalid):
                        pass
                    break
            else:
                self.log.info(f"点击失败: 未找到匹配的按钮文本 {result} {spec}.")
        except KeyError:
            self.log.info(f"点击失败: {result} 不是可用的答案 {spec}.")
        except RPCError:
            self.log.info(f"点击失败: 问题已失效.")


class PornembyAnswerMonitor:
    class PornembyAnswerResultMonitor(_PornembyAnswerResultMonitor):
        chat_name = "Pornemby"

    class PornembyAnswerAnswerMonitor(_PornembyAnswerAnswerMonitor):
        chat_name = "Pornemby"



from json import JSONDecodeError
from urllib.parse import parse_qs, urlparse

from curl_cffi.requests import AsyncSession, RequestsError, Response
from pyrogram.raw.functions.users import GetFullUser
from pyrogram.raw.functions.messages import RequestWebView

from embykeeper.runinfo import RunStatus
from embykeeper.utils import get_proxy_str, show_exception
from embykeeper.config import config

from . import BotCheckin


class AuroraCheckin(BotCheckin):
    name = "Aurora Media"
    bot_username = "AuroraMedia1_bot"
    max_retries = 1
    additional_auth = ["prime"]

    async def send_checkin(self, **kw):
        bot_peer = await self.client.resolve_peer(self.bot_username)
        user_full = await self.client.invoke(GetFullUser(id=bot_peer))
        url = user_full.full_user.bot_info.menu_button.url
        url_auth = (
            await self.client.invoke(RequestWebView(peer=bot_peer, bot=bot_peer, platform="ios", url=url))
        ).url

        scheme = urlparse(url_auth)
        params = parse_qs(scheme.fragment)
        webapp_data = params.get("tgWebAppData", [""])[0]

        # 新的API端点
        base_url = "https://server.auroramedia.me"
        url_login = f"{base_url}/aurora/v1/user/login"
        url_checkin = f"{base_url}/aurora/v1/user/checkin"

        try:
            async with AsyncSession(
                proxy=get_proxy_str(config.proxy), impersonate="edge", allow_redirects=True
            ) as session:
                # 登录获取token
                headers = {"Authorization": f"tma {webapp_data}"}
                resp_login: Response = await session.post(url_login, headers=headers)
                login_results = resp_login.json()

                if login_results.get("code") != 200:
                    self.log.info("登录失败: 账户错误.")
                    return await self.fail(message="账户错误")

                token = login_results["data"]["token"]

                # 使用token进行签到
                headers = {"Authorization": f"Bearer-{token}"}
                resp = await session.get(url_checkin, headers=headers)
                results = resp.json()

                code = results.get("code")
                message = results.get("message")
                if "已经签到过了" in message:
                    self.log.info(f"今日已经签到过了.")
                    return await self.finish(RunStatus.NONEED, "今日已签到")
                if code != 200:
                    self.log.info(f"签到失败, 请求状态为 {code}: {message}.")
                    return await self.retry()

                points = results["data"]["points"]
                add_points = results["data"]["addPoints"]
                self.log.info(f"[yellow]签到成功[/]: + {add_points} 分 -> {points} 分.")
                return await self.finish(RunStatus.SUCCESS, "签到成功")

        except (RequestsError, OSError, JSONDecodeError) as e:
            self.log.info(f"签到失败: 无法连接签到页面 ({e.__class__.__name__}).")
            show_exception(e)
            return await self.retry()



from __future__ import annotations

import asyncio
from typing import Dict, List, Set, Type

from loguru import logger

from embykeeper.schema import TelegramAccount
from embykeeper.config import config
from embykeeper.runinfo import RunContext

from .monitor import Monitor
from .dynamic import extract, get_cls, get_names
from .link import Link
from .session import ClientsSession
from .pyrogram import Client

logger = logger.bind(scheme="telechecker")


class MonitorManager:
    """监控管理器"""

    def __init__(self):
        self._tasks: Dict[str, asyncio.Task] = {}  # phone -> task
        self._running: Set[str] = set()  # Currently running phones

        # Set up config change callbacks
        config.on_list_change("telegram.account", self._handle_account_change)

    def _handle_account_change(self, added: List[TelegramAccount], removed: List[TelegramAccount]):
        """Handle account additions and removals"""
        for account in removed:
            logger.info(f"{account.phone} 账号的群组监控任务已被清除.")
            self.stop_account(account.phone)

        for account in added:
            if account.monitor and account.enabled:
                logger.info(f"新增的 {account.phone} 账号的群组监控任务已增加.")
                self.start_account(account)

    def stop_account(self, phone: str):
        """Stop running tasks for an account"""
        if phone in self._tasks:
            self._tasks[phone].cancel()
            del self._tasks[phone]

        self._running.discard(phone)

    def start_account(self, account: TelegramAccount):
        """Start monitoring for an account"""
        if not account.monitor or account.phone in self._running:
            return
        task = asyncio.create_task(self.run_account(account))
        self._tasks[account.phone] = task
        return task

    async def run_account(self, account: TelegramAccount):
        """Run monitor for a single account"""
        if account.phone in self._running:
            logger.warning(f"账户 {account.phone} 的监控已经在执行.")
            return

        account_ctx = RunContext.get_or_create(f"monitor.account.{account.phone}")

        self._running.add(account.phone)
        try:
            async with ClientsSession([account]) as clients:
                async for a, client in clients:
                    await RunContext.run(
                        lambda c: self._run_account(c, a, client),
                        description=f"{account.phone} 账号监控",
                        parent_ids=[account_ctx.id],
                    )
        finally:
            self._running.discard(account.phone)

    async def _run_account(self, ctx: RunContext, account: TelegramAccount, client: Client):
        """Run monitors for a single user"""
        log = logger.bind(username=client.me.name)

        # Get monitor classes based on account config or global config
        site = None
        if account.site and account.site.monitor is not None:
            site = account.site.monitor
        elif config.site and config.site.monitor is not None:
            site = config.site.monitor
        else:
            site = get_names("monitor")

        clses: List[Type[Monitor]] = extract(get_cls("monitor", names=site))

        if not clses:
            if site is not None:  # Only show warning if sites were specified but none were valid
                log.warning("没有任何有效监控站点, 监控将跳过.")
            return

        if not await Link(client).auth("monitor", log_func=log.error):
            return

        monitors = []
        names = []

        for cls in clses:
            if hasattr(cls, "templ_name"):
                site_name = cls.templ_name
            else:
                site_name = cls.__module__.rsplit(".", 1)[-1]
            site_ctx = RunContext.prepare(f"{site_name} 站点监控", parent_ids=ctx.id)
            monitor = cls(
                client,
                context=site_ctx,
                config=config.monitor.get_site_config(site_name),
            )
            monitors.append(monitor)
            names.append(monitor.name)

        if names:
            log.debug(f'已启用监控器: {", ".join(names)}')

        # Start all monitors concurrently
        await asyncio.gather(*[m._start() for m in monitors])

    async def run_all(self):
        """Run monitoring for all enabled accounts"""
        accounts = [a for a in config.telegram.account if a.enabled and a.monitor]
        tasks = []
        for account in accounts:
            task = self.start_account(account)
            if task:  # start_account might return None if account is already running
                tasks.append(task)
        if tasks:
            await asyncio.gather(*tasks)



from .follow import FollowMonitor

__ignore__ = True


class TestFollowMonitor(FollowMonitor):
    name = "全部群组从众 测试"
    chat_name = "api_group"
    chat_allow_outgoing = True
    chat_follow_user = 3
    allow_same_user = True



import asyncio
from datetime import datetime, timedelta
import random
import re
from urllib.parse import parse_qs, urlparse

from pyrogram.types import Message
from pyrogram.raw.functions.messages import RequestWebView
from pyrogram.errors import MessageIdInvalid
from faker import Faker
import httpx

from embykeeper.config import config
from embykeeper.utils import show_exception, truncate_str, get_proxy_str

from ..link import Link
from . import BotCheckin


class FutureCheckin(BotCheckin):
    name = "未响"
    bot_username = "lotayu_bot"
    bot_use_captcha = False
    bot_checkin_cmd = "/start"
    bot_text_ignore = ["請先完成驗證"]
    bot_fail_keywords = ["不能签到"]
    additional_auth = ["captcha"]
    max_retries = 2

    click_button = ["签到", "簽到"]

    async def send_checkin(self, retry=False):
        """发送签到命令, 或依次发送签到命令序列."""
        history_message = await self.get_history_message(limit=10)
        if history_message:
            await self.message_handler(self.client, history_message)
            return
        return await super().send_checkin(retry=retry)

    async def get_history_message(self, limit=0):
        """处理 limit 条历史消息, 并检测是否有验证."""
        try:
            m: Message
            async for m in self.client.get_chat_history(self.chat_name or self.bot_username, limit=limit):
                if m.text and "點擊下方按鈕並驗證您的身份" in m.text:
                    time_match = re.search(r"當前時間:(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})", m.text)
                    if not time_match:
                        return None
                    time_str = time_match.group(1)
                    current_time = datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S")
                    now = datetime.now()
                    time_difference = now - current_time
                    if time_difference <= timedelta(minutes=2):
                        return m
                    else:
                        return None
                elif m.caption and "您的驗證已經通過" in m.caption:
                    return None
        except Exception as e:
            self.log.warning("读取历史消息失败, 将不再读取历史消息.")
            show_exception(e)
            return None

    async def message_handler(self, client, message: Message):
        if message.text and "未加入" in message.text:
            self.log.warning(f"签到失败: 账户错误.")
            return await self.fail()

        if message.text and "您有一個還在進行中的驗證會話" in message.text:
            self.log.warning(f"签到失败: 验证码解析异常, 之前有未完成的验证.")
            return await self.fail()

        if message.text and "驗證您的身份" in message.text and message.reply_markup:
            keys = [b for r in message.reply_markup.inline_keyboard for b in r]
            for b in keys:
                if "Verify" in b.text and b.web_app:
                    url = b.web_app.url
                    bot_peer = await self.client.resolve_peer(self.bot_username)
                    url_auth = (
                        await self.client.invoke(
                            RequestWebView(peer=bot_peer, bot=bot_peer, platform="ios", url=url)
                        )
                    ).url
                    if not await self.solve_captcha(url_auth):
                        self.log.warning("签到失败: 验证码解析失败, 正在重试.")
                        await asyncio.sleep(self.bot_retry_wait)
                        await self.retry()
                        return
                    else:
                        await asyncio.sleep(random.uniform(3, 5))
                        self.log.info("已成功验证, 即将重新进行签到流程.")
                        return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()

        if (
            message.caption
            and ("開始回響" in message.caption or "開始操作" in message.caption)
            and message.reply_markup
        ):
            keys = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            for k in keys:
                if any([i in k for i in self.click_button]):
                    try:
                        await message.click(k)
                    except (MessageIdInvalid, TimeoutError):
                        pass
                    return
            else:
                self.log.warning(f"签到失败: 账户错误.")
                return await self.fail()
        await super().message_handler(client, message)

    async def solve_captcha(self, url: str):
        token = await Link(self.client).captcha("future_echo")
        if not token:
            return False
        else:
            scheme = urlparse(url)
            params = parse_qs(scheme.query)
            url_submit = scheme._replace(path="/x/api/submit", query="", fragment="").geturl()
            uuid = params.get("id", [None])[0]
            origin = scheme._replace(path="/", query="", fragment="").geturl()
            useragent = Faker().safari()
            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "Referer": url,
                "Origin": origin,
                "User-Agent": useragent,
            }
            data = {
                "uuid": uuid,
                "cf-turnstile-response": token,
            }
            for i in range(10):
                try:
                    async with httpx.AsyncClient(http2=True, proxy=get_proxy_str(config.proxy)) as client:
                        resp = await client.post(url_submit, headers=headers, data=data)
                        result = resp.text
                        if "完成" in result:
                            return True
                        else:
                            self.log.warning(
                                f"验证码识别后接口返回异常信息:\n{truncate_str(result, 100)}, 可能是您的请求 IP 风控等级较高导致的."
                            )
                            return False
                except (httpx.ProxyError, httpx.TimeoutException, OSError):
                    self.log.warning(
                        f"无法连接到站点的页面, 可能是您的网络或代理不稳定, 正在重试 ({i+1}/10)."
                    )
                    continue
            else:
                self.log.warning(f'无法连接到站点的页面: "{url_submit}", 可能是您的网络或代理不稳定.')
                return False



import os
import platform
import sys
from msvcrt import getch
from subprocess import Popen

from rich import box
from rich.align import Align
from rich.panel import Panel
from rich.table import Table

from . import __name__ as __product__
from . import var
from .cli import app as cli
from .config import config


def generate_config():
    config_file = config.basedir / "config.toml"
    if config_file.exists():
        return
    try:
        with open(config_file, "w+", encoding="utf-8") as f:
            f.write(config.generate_example_config())
    except OSError as e:
        var.console.print(
            f'无法写入默认配置文件 "{config_file}", 请确认是否有权限进行该目录写入: {e}, 请按任意键退出',
            justify="center",
        )
        _ = getch()
        sys.exit(1)
    message = Table.grid(padding=2)
    message.add_column()
    urls = Table.grid(padding=2)
    urls.add_column(style="green", justify="right")
    urls.add_column(no_wrap=True)
    urls.add_row("配置教程", "https://emby-keeper.github.io/guide/配置文件")
    message.add_row("您需要在即将打开的文件中进行账户配置, 您可以根据注释提示进行配置.")
    message.add_row("您也可以参考 Wiki, 链接如下:")
    message.add_row(urls)
    message.add_row(Align("请按任意键以打开模板配置文件", align="center", style="bold blink underline"))
    var.console.print(
        Panel.fit(
            message,
            box=box.ROUNDED,
            padding=(1, 2),
            title="[b red]欢迎您使用 Embykeeper!",
            border_style="bright_blue",
        ),
        justify="center",
    )

    _ = getch()

    p = Popen(["notepad.exe", str(config_file)], shell=True)

    if not p:
        var.console.print(f"配置文件打开失败, 请按任意键退出", justify="center")
        _ = getch()
        sys.exit(1)

    var.console.print(f"等待您编辑配置文件, 请保存关闭编辑器窗口以继续 ...", justify="center")
    p.wait()
    var.console.print(
        f"请确认您配置完成, 并按任意键以继续启动 {__product__.capitalize()}...", justify="center"
    )
    var.console.print(f"配置完成, 即将启动 {__product__.capitalize()} ...", justify="center")


def main():
    generate_config()
    os.system("cls")
    var.console.rule("Embykeeper")
    args = sys.argv[1:]
    if "-I" in args:
        args = sys.argv[1:] + ["-W"]
    else:
        args = sys.argv[1:] + ["-W", "-i"]
    cli(args)


if __name__ == "__main__":
    if not platform.system() == "Windows":
        raise RuntimeError("this entrypoint is only for windows")
    main()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Dict, Tuple, Union
from asyncio import Task

import psutil
from rich.live import Live
from rich.table import Table
from rich.rule import Rule
from rich.console import Group

from .telegram.session import ClientsSession

from .var import console, tele_used, emby_used

if TYPE_CHECKING:
    from .telegram.pyrogram import Client


async def topper():
    """在控制台底部实时显示系统资源使用情况."""

    process = psutil.Process()

    def get_client_stats(pool: Dict[str, Tuple[Union[Client, Task], int]]) -> Tuple[int, int, int, str]:
        """统计 Client 状态数量"""
        failed = pending = using = idle = 0
        queue_stats = []
        for v in pool.values():
            if not v:
                failed += 1
            if isinstance(v, Task):
                pending += 1
            else:
                client, count = v
                if count > 0:
                    using += 1
                else:
                    idle += 1
                # 获取队列和任务统计
                if hasattr(client, "dispatcher"):
                    try:
                        qsize = client.dispatcher.updates_queue.qsize()
                        tasks = client.dispatcher.handler_worker_tasks
                        active = sum(1 for t in tasks if t.get_coro().cr_await.__name__ != "get")
                        if qsize > 0 or active > 0:
                            # 当队列超过10或handler使用率超过80%时显示红色
                            if qsize >= 10 or (active / len(tasks) >= 0.8):
                                queue_stats.append(f"[red][{qsize}:{active}/{len(tasks)}][/red]")
                            else:
                                queue_stats.append(f"[{qsize}:{active}/{len(tasks)}]")
                    except:
                        queue_stats.append("[Error]")

        queue_text = f" Queue: {' '.join(queue_stats)}" if queue_stats else ""
        return failed, pending, using, idle, queue_text

    def get_ocr_stats():
        """获取OCR子进程状态"""
        children = process.children()
        if not children:
            return None
        total_mem = sum(p.memory_info().rss for p in children) / 1024 / 1024
        return f"OCR: {len(children)} ({total_mem:.1f} MB)"

    def get_stats():
        # 创建状态表格
        table = Table(show_header=False, box=None)

        # 系统资源状态
        mem_mb = process.memory_info().rss / 1024 / 1024
        cpu_percent = process.cpu_percent()
        mem_text = f"[red]{mem_mb:.1f}[/red]" if mem_mb > 1024 else f"{mem_mb:.1f}"
        sys_stats = [
            ("Embykeeper >", "bright_blue"),
            (f"MEM: {mem_text} MB, CPU: {cpu_percent:.1f}%", "bright_blue"),
        ]

        # OCR状态
        ocr_stats = get_ocr_stats()
        if ocr_stats:
            sys_stats.append((ocr_stats, "bright_blue"))

        # Client状态
        if tele_used:
            from .telegram.pyrogram import Dispatcher
            from .telegram.link import Link

            failed, pending, using, idle, queue_text = get_client_stats(ClientsSession.pool)
            client_stats = []
            if failed:
                client_stats.append(f"Failed({failed})")
            if pending:
                client_stats.append(f"Pending({pending})")
            if using:
                client_stats.append(f"Using({using})")
            if idle:
                client_stats.append(f"Idle({idle})")
            if client_stats:
                sys_stats.append((f"Tele: {'/'.join(client_stats)}{queue_text}", "bright_blue"))

            if Link.post_count > 0:
                sys_stats.append((f"Link: {Link.post_count}", "bright_blue"))

            if Dispatcher.updates_count > 0:
                sys_stats.append((f"Updates: {Dispatcher.updates_count}", "bright_blue"))

        if emby_used:
            from .emby.api import Emby

            if Emby.playing_count > 0:
                sys_stats.append((f"Play: {Emby.playing_count}", "bright_blue"))

        table.add_column(style="bright_blue", justify="left")
        for _ in range(len(sys_stats) - 1):
            table.add_column(style="bright_blue", justify="left")

        table.add_row(*[text for text, _ in sys_stats])
        return Group(Rule(style="bright_blue"), table)

    live = Live(
        console=console,
        refresh_per_second=1,
        vertical_overflow="visible",
        auto_refresh=True,
    )
    try:
        live.start()
        while True:
            live.update(get_stats())
            await asyncio.sleep(1)
    except (KeyboardInterrupt, asyncio.CancelledError):
        live.update("")  # 先清空显示内容
        live.stop()  # 然后停止 Live
        raise



from . import BotCheckin

__ignore__ = True


class FanhuaSGKCheckin(BotCheckin):
    name = "繁花社工库"
    bot_username = "FanHuaSGK_bot"
    bot_checkin_cmd = "/sign"
    bot_checked_keywords = ["今日已签到"]



import random
from typing import Iterable, List, Optional, Union
from loguru import logger
from pydantic import BaseModel, ValidationError

from embykeeper.ocr import CharRange
from embykeeper.utils import to_iterable

from . import BotCheckin

__ignore__ = True


class TemplateBCheckinConfig(BaseModel):
    # fmt: off
    name: str = None  # 签到器的名称
    bot_checkin_cmd: Union[str, List[str]] = ["/checkin"]  # Bot 依次执行的签到命令
    bot_send_interval: int = 3  # 签到命令间等待的秒数
    bot_use_captcha: bool = True  # 当 Bot 返回图片时, 识别验证码并调用 on_captcha
    bot_checkin_caption_pat: Optional[str] = (None)  # 当 Bot 返回图片时, 仅当符合该 regex 才识别为验证码, 置空不限制
    bot_text_ignore: Union[str, List[str]] = []  # 当含有列表中的关键词, 即忽略该消息, 置空不限制
    ocr: Optional[str] = None  # OCR 模型, None = 默认模型, str = 自定义模型
    bot_captcha_char_range: Optional[Union[CharRange, str]] = (None)  # OCR 字符范围, 仅当默认模型可用, None = 默认范围, OCRRanges = 预定义范围, str = 自定义范围
    bot_captcha_len: Union[int, Iterable[int]] = []  # 验证码长度的可能范围, 例如 [1, 2, 3], 置空不限制
    bot_success_pat: str = r"(\d+)[^\d]*(\d+)"  # 当接收到成功消息后, 从消息中提取数字的模式
    bot_retry_wait: int = 2  # 失败时等待的秒数
    bot_use_history: Optional[int] = None  # 首先尝试识别历史记录中最后一个验证码图片, 最多识别 N 条, 置空禁用
    bot_allow_from_scratch: bool = False  # 允许从未聊天情况下启动
    bot_success_keywords: Union[str, List[str]] = ([])  # 成功时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_checked_keywords: Union[str, List[str]] = []  # 今日已签到时检测的关键词, 置空使用内置关键词表
    bot_account_fail_keywords: Union[str, List[str]] = ([])  # 账户错误将退出时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_too_many_tries_fail_keywords: Union[str, List[str]] = ([])  # 过多尝试将退出时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_fail_keywords: Union[str, List[str]] = ([])  # 签到错误将重试时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    is_chat: bool = False  # 指定的用户名为群组用户名或 ID, 而非机器人
    max_retries: Optional[int] = None  # 验证码错误或网络错误时最高重试次数 (默认无限)
    checked_retries: Optional[int] = None  # 今日已签到时最高重试次数 (默认不重试)
    wait_response: bool = True  # 是否需要等待相关回复, 以确认签到完成
    # fmt: on


class TemplateBCheckin(BotCheckin):
    init_first = True
    additional_auth = ["prime"]

    async def init(self):
        try:
            self.t_config = TemplateBCheckinConfig.model_validate(self.config)
        except ValidationError as e:
            self.log.warning(f"初始化失败: 签到自定义模板 B 的配置错误:\n{e}")
            return False
        self.name = self.t_config.name or "自定义"

        if self.t_config.is_chat:
            self.chat_name = self.bot_username
            self.bot_username = None
        self.bot_checkin_cmd = self.t_config.bot_checkin_cmd
        self.bot_send_interval = self.t_config.bot_send_interval
        self.bot_use_captcha = self.t_config.bot_use_captcha
        self.bot_checkin_caption_pat = self.t_config.bot_checkin_caption_pat
        self.bot_text_ignore = self.t_config.bot_text_ignore
        self.ocr = self.t_config.ocr
        self.bot_captcha_char_range = self.t_config.bot_captcha_char_range
        self.bot_captcha_len = self.t_config.bot_captcha_len
        self.bot_success_pat = self.t_config.bot_success_pat
        self.bot_retry_wait = self.t_config.bot_retry_wait
        self.bot_use_history = self.t_config.bot_use_history
        self.bot_allow_from_scratch = self.t_config.bot_allow_from_scratch
        self.bot_success_keywords = self.t_config.bot_success_keywords
        self.bot_checked_keywords = self.t_config.bot_checked_keywords
        self.bot_account_fail_keywords = self.t_config.bot_account_fail_keywords
        self.bot_too_many_tries_fail_keywords = self.t_config.bot_too_many_tries_fail_keywords
        self.bot_fail_keywords = self.t_config.bot_fail_keywords
        self.max_retries = self.t_config.max_retries
        self.checked_retries = self.t_config.checked_retries

        self.log = logger.bind(scheme="telechecker", name=self.name, username=self.client.me.name)
        return True

    async def send_checkin(self, **kw):
        if self.chat_name:
            cmd = random.choice(to_iterable(self.bot_checkin_cmd))
            await self.send(cmd)
            if not self.t_config.wait_response:
                await self.finish(message="群组发言已发送")
        else:
            await super().send_checkin(**kw)
            if not self.t_config.wait_response:
                await self.finish(message="签到命令已发送")


def use(**kw):
    return type("TemplatedClass", (TemplateBCheckin,), kw)



from .pornemby_answer import _PornembyAnswerAnswerMonitor, _PornembyAnswerResultMonitor

__ignore__ = True


class TestPornembyAnswerMonitor:
    class _TestPornembyAnswerAnswerMonitor(_PornembyAnswerAnswerMonitor):
        name = "Pornemby 问题回答测试"
        chat_name = "api_group"
        chat_user = ["embykeeper_test_bot"]

    class _TestPornembyAnswerResultMonitor(_PornembyAnswerResultMonitor):
        name = "Pornemby 问题答案测试"
        chat_name = "api_group"
        chat_user = ["embykeeper_test_bot"]



import random
from embykeeper.utils import to_iterable
from . import BotCheckin

__ignore__ = True


class FeiyueMusicGroupCheckin(BotCheckin):
    name = "飞跃星空群组发言"
    chat_name = -1002197507537
    bot_checkin_cmd = "签到"
    skip = 14

    async def send_checkin(self, **kw):
        cmd = random.choice(to_iterable(self.bot_checkin_cmd))
        await self.send(cmd)
        await self.finish(message="群组发言已发送")



from .pornemby_double import PornembyDoubleMonitor

__ignore__ = True


class TestPornembyDoubleMonitor(PornembyDoubleMonitor):
    name = "Pornemby 怪兽自动翻倍 测试"
    chat_name = "api_group"
    chat_user = "embykeeper_test_bot"
    chat_allow_outgoing = True



from pyrogram.types import Message

from . import Monitor

__ignore__ = True


class TestPornembyMonitor(Monitor):
    name = "Pornemby 消息接收 测试"
    chat_name = "Pornemby"
    chat_keyword = r".*"

    async def on_trigger(self, message: Message, key, reply):
        self.log.info(f"Pornemby 消息接收 测试: {message.text or message.caption}")



from rich.prompt import Prompt

from .cache import cache
from .var import console


def get_cache_options():
    """获取缓存清理选项"""
    return {
        "1": {"name": "除凭据外所有缓存", "special": "all_except_credentials"},
        "2": {"name": "Emby 模拟设备信息", "prefix": "emby.env"},
        "3": {
            "name": "运行任务日志记录信息",
            "prefix": "runinfo",
            "extra_delete": lambda key: f"runinfo.children.{key.split('.')[-1]}",
        },
        "4": {"name": "下次运行时间缓存", "prefix": "scheduler"},
        "5": {"name": "Emby 登陆凭据", "prefix": "emby.credential", "show_keys": True},
        "6": {"name": "Telegram 登陆凭据", "prefix": "telegram.session_str", "show_keys": True},
        "7": {
            "name": "其他缓存",
            "children": {
                "7.1": {"name": "monitor.pornemby.answer.qa", "prefix": "monitor.pornemby.answer.qa"}
            },
        },
    }


def clean_cache(cache_key: str = None, cache_prefix: str = None):
    """清理指定的缓存

    Args:
        cache_key: 具体的缓存键
        cache_prefix: 缓存前缀
    """
    if cache_key:
        # 清理单个缓存键
        cache.delete(cache_key)
        return f"已清理缓存 {cache_key}"

    if cache_prefix is not None:
        # 清理指定前缀的所有缓存
        if cache_prefix == "all_except_credentials":
            # 特殊处理：清理除凭据外所有缓存
            credential_prefixes = ["emby.credential", "telegram.session_str"]
            all_keys = cache.find_by_prefix("")
            keys_to_delete = [k for k in all_keys if not any(k.startswith(p) for p in credential_prefixes)]
            count = len(keys_to_delete)
            cache.delete_many(keys_to_delete)
            return f"已清理除凭据外所有缓存，共 {count} 条"
        else:
            # 常规前缀清理
            keys = cache.find_by_prefix(cache_prefix)
            count = len(keys)
            cache.delete_many(keys)
            return f"已清理前缀为 {cache_prefix} 的缓存，共 {count} 条"

    return "请指定要清理的缓存键或前缀"


async def cleaner():
    options = get_cache_options()
    console.rule("缓存文件清理")
    console.print("可用的清理选项：")

    for key, option in options.items():
        if "prefix" in option:
            keys = list(cache.find_by_prefix(option["prefix"]))
            count = len(keys)
            if option.get("show_keys", False):
                # 显示凭据类型，需要显示具体的key
                console.print(f"{key}. {option['name']} ({option['prefix']})")
                for i, key_name in enumerate(keys, 1):
                    console.print(f"  {key}.{i}. {key_name}")
            else:
                # 显示普通缓存类型，显示总数
                console.print(f"{key}. {option['name']} (共 {count} 条)")
        elif "children" in option:
            # 显示父选项（如"其他缓存"）
            console.print(f"{key}. {option['name']}")
            for child_key, child in option["children"].items():
                keys = list(cache.find_by_prefix(child["prefix"]))
                count = len(keys)
                # 显示子选项
                console.print(f"  {child_key}. {child['name']} (共 {count} 条)")
                for i, key_name in enumerate(keys, 1):
                    console.print(f"    {child_key}.{i}. {key_name}")
        else:
            # 特殊选项（如"除凭据外所有缓存"）
            console.print(f"{key}. {option['name']}")

    console.rule()
    option = Prompt.ask("\n请输入要清理的选项编号")

    # 根据选项获取实际的缓存键或前缀
    parts = option.split(".", 1)
    parent_key = parts[0]

    if parent_key not in options:
        result = f"无效的选项: {parent_key}"
    else:
        target = options[parent_key]
        if "special" in target:
            result = clean_cache(cache_prefix=target["special"])
        elif "prefix" in target:
            if target.get("show_keys", False) and len(parts) > 1:
                # 用户选择了具体的凭据
                keys = list(cache.find_by_prefix(target["prefix"]))
                try:
                    index = int(parts[1]) - 1
                    if 0 <= index < len(keys):
                        key_to_clean = keys[index]
                        result = clean_cache(cache_key=key_to_clean)
                    else:
                        result = "无效的选项索引"
                except ValueError:
                    result = "无效的选项格式"
            else:
                # 清理整个前缀
                result = clean_cache(cache_prefix=target["prefix"])
        elif "children" in target:
            if len(parts) > 1:
                # 用户输入了子选项，如 "7.1"
                child_key = parts[1].split(".", 1)[0]
                full_child_key = f"{parent_key}.{child_key}"

                if full_child_key in target["children"]:
                    child = target["children"][full_child_key]
                    result = clean_cache(cache_prefix=child["prefix"])
                else:
                    result = f"无效的子选项: {full_child_key}"
            else:
                # 用户只输入了父选项，如 "7"
                # 清理所有子选项
                results = []
                for child_key, child in target["children"].items():
                    prefix = child["prefix"]
                    res = clean_cache(cache_prefix=prefix)
                    results.append(f"{child_key}: {res}")

                if results:
                    result = "\n".join(results)
                else:
                    result = "没有可清理的子选项"
        else:
            result = "无效的选项类型"

    console.print(result + "\n")
    console.rule()



from __future__ import annotations

import inspect
import pkgutil
import re
from functools import lru_cache
from importlib import import_module
from typing import List, Type

from loguru import logger

from embykeeper.utils import to_iterable

from . import __name__ as __telechecker__

logger = logger.bind(scheme="telegram")


def get_spec(type: str) -> tuple[str, str]:
    """模块路径解析映射."""
    if type == "checkiner":
        sub = "checkiner"
        suffix = "checkin"
    elif type == "monitor":
        sub = "monitor"
        suffix = "monitor"
    elif type == "messager":
        sub = "messager"
        suffix = "messager"
    else:
        raise ValueError(f"{type} is not a valid service.")
    return sub, suffix


@lru_cache
def get_names(type: str, allow_ignore=False) -> List[str]:
    """列出服务中所有可用站点."""
    sub, _ = get_spec(type)
    results = []
    typemodule = import_module(f"{__telechecker__}.{sub}")
    for _, mn, _ in pkgutil.iter_modules(typemodule.__path__):
        module = import_module(f"{__telechecker__}.{sub}.{mn}")
        if not allow_ignore:
            if not getattr(module, "__ignore__", False):
                results.append(mn)
        else:
            if (not mn.startswith("_")) and (not mn.startswith("test")):
                results.append(mn)
    return results


def get_cls(type: str, names: List[str] = None) -> List[Type]:
    """获得服务特定站点的所有类."""

    sub, suffix = get_spec(type)
    if names == None:
        names = get_names(type)
    else:
        names = to_iterable(names)

    exclude_names = set(name[1:] for name in names if name.startswith("-"))
    include_names = set(name[1:] for name in names if name.startswith("+"))
    names = set(name for name in names if not name.startswith("-") and not name.startswith("+"))

    if not names and (exclude_names or include_names):
        names = set(get_names(type))

    if "all" in names:
        names = set(get_names(type, allow_ignore=True))

    if type == "checkiner":
        if "sgk" in names:
            sgk_names = set(n for n in get_names(type, allow_ignore=True) if n.endswith("sgk"))
            names.update(sgk_names)
            names.remove("sgk")

        if "sgk" in exclude_names:
            sgk_names = set(n for n in names if n.endswith("sgk"))
            names -= sgk_names
            exclude_names.remove("sgk")

        if "sgk" in include_names:
            sgk_names = set(n for n in get_names(type, allow_ignore=True) if n.endswith("sgk"))
            include_names.update(sgk_names)
            include_names.remove("sgk")

    # Apply exclusions
    names = names - exclude_names
    # Add inclusions
    names = list(names | include_names)

    results = []
    for name in names:
        match = re.match(r"templ_(\w+)<(\w+)>", name)
        if match:
            try:
                module = import_module(f"{__telechecker__}.{sub}._templ_{match.group(1).lower()}")
                func = getattr(module, "use", None)
                if not func:
                    logger.warning(f'您配置的 "{type}" 不支持模板 "{match.group(1).upper()}".')
                    continue
                if type == "checkiner":
                    results.append(
                        func(bot_username=match.group(2), name=f"@{match.group(2)}", templ_name=name)
                    )
                elif type == "monitor":
                    results.append(func(name=f"{match.group(2)}", templ_name=name))
                elif type == "messager":
                    results.append(func(chat_name=match.group(2), name=f"@{match.group(2)}", templ_name=name))
            except ImportError:
                all_names = get_names(type)
                logger.warning(f'您配置的 "{type}" 不支持站点 "{name}", 请从以下站点中选择:')
                logger.warning(", ".join(all_names))
        else:
            try:
                module = import_module(f"{__telechecker__}.{sub}.{name.lower()}")
                found_valid_class = False
                expected_name = name.replace("_old", "").replace("_", "")
                for cn, cls in inspect.getmembers(module, inspect.isclass):
                    if (expected_name + suffix).lower() == cn.lower():
                        results.append(cls)
                        found_valid_class = True
                if not found_valid_class:
                    logger.warning(
                        f'您设定了站点 "{name}", 但对应的模块 "{name}" 中未找到合法的类, '
                        f'类名应以 "{expected_name.capitalize() + suffix.capitalize()}" 开头.'
                    )
            except ImportError:
                all_names = get_names(type)
                logger.warning(f'您配置的 "{type}" 不支持站点 "{name}", 请从以下站点中选择:')
                logger.warning(", ".join(all_names))
    return results


def extract(clss: List[Type]) -> List[Type]:
    """对于嵌套类, 展开所有子类."""
    extracted = []
    for cls in clss:
        ncs = [c for c in cls.__dict__.values() if inspect.isclass(c)]
        if ncs:
            extracted.extend(ncs)
        else:
            extracted.append(cls)
    return extracted



import asyncio
from datetime import datetime, timedelta
import random
import re
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from enum import Flag, auto
import string
import time
from typing import Iterable, List, Optional, Union

from loguru import logger
from pyrogram import filters
from pyrogram.errors import (
    UsernameNotOccupied,
    FloodWait,
    UsernameInvalid,
    ChannelInvalid,
    ChannelPrivate,
    MessageIdInvalid,
    DataInvalid,
)
from pyrogram.handlers import EditedMessageHandler, MessageHandler
from pyrogram.types import InlineKeyboardMarkup, Message, ReplyKeyboardMarkup
from pyrogram.raw.functions.account import GetNotifySettings
from pyrogram.raw.types import PeerNotifySettings, InputNotifyPeer
from thefuzz import fuzz, process

from embykeeper import __name__ as __product__
from embykeeper.ocr import CharRange, OCRService
from embykeeper.runinfo import RunContext
from embykeeper.utils import show_exception, to_iterable, format_timedelta_human, AsyncCountPool
from embykeeper.config import config
from embykeeper.runinfo import RunStatus
from embykeeper.telegram.pyrogram import Client
from embykeeper.telegram.link import Link

__ignore__ = True

logger = logger.bind(scheme="telechecker")

default_keywords = {
    "account_fail": (
        "拉黑",
        "黑名单",
        "冻结",
        "未找到用户",
        "无资格",
        "退出群",
        "退群",
        "加群",
        "加入群聊",
        "请先关注",
        "请先加入",
        "請先加入",
        "未注册",
        "先注册",
        "不存在",
        "不在群组中",
    ),
    "too_many_tries_fail": ("已尝试", "过多"),
    "checked": ("只能", "已经", "下次", "过了", "签过", "明日再来", "上次签到", "重复签到"),
    "fail": ("失败", "错误", "超时"),
    "success": ("成功", "通过", "完成", "获得"),
}


class MessageType(Flag):
    IGNORE = auto()
    TEXT = auto()
    CAPTION = auto()
    CAPTCHA = auto()
    ANSWER = auto()


class BaseBotCheckin(ABC):
    """基础签到类."""

    name: str = None

    def __init__(
        self,
        client: Client,
        context: RunContext = None,
        retries=None,
        timeout=None,
        config: dict = {},
    ):
        """
        基础签到类.
        参数:
            client: Pyrogram 客户端
            context: 运行时上下文
            retries: 最大重试次数
            timeout: 签到超时时间
            config: 当前签到器的特定配置
        """
        self.client = client
        self.ctx = context or RunContext.prepare()

        self._retries = retries
        self._timeout = timeout

        self.config = config
        self.finished = asyncio.Event()  # 签到完成事件
        self.log = self.ctx.bind_logger(logger.bind(name=self.name, username=client.me.name))  # 日志组件

        self._task = None  # 主任务

    @property
    def retries(self):
        return self._retries or config.checkiner.retries

    @property
    def timeout(self):
        return self._timeout or config.checkiner.timeout

    async def _start(self):
        """签到器的入口函数的错误处理外壳."""
        try:
            self.client.stop_handlers.append(self.stop)
            self._task = asyncio.create_task(self.start())
            return await self._task
        except Exception as e:
            if config.nofail:
                self.log.warning(f"初始化异常错误, 签到器将停止.")
                show_exception(e, regular=False)
                return self.ctx.finish(RunStatus.ERROR, "异常错误")
            else:
                raise
        finally:
            self.client.stop_handlers.remove(self.stop)
            self._task = None

    @abstractmethod
    async def start(self) -> RunContext:
        pass

    async def stop(self):
        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass


class BotCheckin(BaseBotCheckin):
    """签到类, 用于回复模式签到."""

    group_pool = AsyncCountPool(base=2000)
    interval_pool = {}

    # fmt: off
    name: str = None  # 签到器的名称
    bot_username: Union[int, str] = None  # Bot 的 UserID 或 用户名 (不带 @ 或 https://t.me/)
    bot_checkin_cmd: Union[str, List[str]] = ["/checkin"]  # Bot 依次执行的签到命令
    bot_send_interval: int = 3  # 签到命令间等待的秒数
    bot_use_captcha = True # 当 Bot 返回图片时, 识别验证码并调用 on_captcha
    bot_checkin_caption_pat: str = None  # 当 Bot 返回图片时, 仅当符合该 regex 才识别为验证码, 置空不限制
    bot_text_ignore: Union[str, List[str]] = []  # 当含有列表中的关键词, 即忽略该消息, 置空不限制
    ocr: Optional[str] = None # OCR 模型, None = 默认模型, str = 自定义模型
    bot_captcha_char_range: Optional[Union[CharRange, str]] = None # OCR 字符范围, 仅当默认模型可用, None = 默认范围, OCRRanges = 预定义范围, str = 自定义范围
    bot_captcha_len: Union[int, Iterable[int]] = []  # 验证码长度的可能范围, 例如 [1, 2, 3], 置空不限制
    bot_success_pat: str = r"(\d+)[^\d]*(\d+)"  # 当接收到成功消息后, 从消息中提取数字的模式
    bot_retry_wait: int = 2  # 失败时等待的秒数
    bot_use_history: int = None  # 首先尝试识别历史记录中最后一个验证码图片, 最多识别 N 条, 置空禁用
    bot_allow_from_scratch: bool = False  # 允许从未聊天情况下启动
    bot_success_keywords: Union[str, List[str]] = []  # 成功时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_checked_keywords: Union[str, List[str]] = []  # 今日已签到时检测的关键词, 置空使用内置关键词表
    bot_account_fail_keywords: Union[str, List[str]] = []  # 账户错误将退出时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_too_many_tries_fail_keywords: Union[str, List[str]] = []  # 过多尝试将退出时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    bot_fail_keywords: Union[str, List[str]] = []  # 签到错误将重试时检测的关键词 (暂不支持regex), 置空使用内置关键词表
    chat_name: str = None  # 在群聊中向机器人签到
    additional_auth: List[str] = []  # 额外认证要求
    max_retries = None  # 验证码错误或网络错误时最高重试次数 (默认无限)
    checked_retries = None # 今日已签到时最高重试次数 (默认不重试)
    init_first: bool = False  # 先执行自定义初始化函数, 再进行加入群组分析
    # fmt: on

    @property
    def valid_retries(self):
        if self.max_retries:
            return min(self.retries, self.max_retries)
        else:
            return self.retries

    def __init__(self, *args, instant=False, **kw):
        super().__init__(*args, **kw)
        self.current_retries = 0  # 当前重试次数
        self._waiting = {}  # 当前等待的消息
        self._handler_tasks = set()  # 存储所有正在运行的 message_handler 任务
        if instant:
            self.checked_retries = None

    def get_filter(self):
        """设定要签到的目标."""
        filter = filters.all
        if self.bot_username:
            filter = filter & filters.user(self.bot_username)
        if self.chat_name:
            filter = filter & filters.chat(self.chat_name)
        else:
            filter = filter & filters.private
        return filter

    def get_handlers(self):
        """设定要监控的更新的类型."""
        return [
            MessageHandler(self._message_handler, self.get_filter()),
            EditedMessageHandler(self._message_handler, self.get_filter()),
        ]

    @asynccontextmanager
    async def listener(self):
        """执行监控上下文."""
        group = await self.group_pool.append(self)
        handlers = self.get_handlers()
        for h in handlers:
            await self.client.add_handler(h, group=group)
        try:
            yield
        finally:
            # 取消所有正在运行的message_handler任务
            for task in self._handler_tasks:
                task.cancel()
            # 等待所有任务完成
            if self._handler_tasks:
                await asyncio.gather(*self._handler_tasks, return_exceptions=True)
            # 移除handlers
            for h in handlers:
                try:
                    await self.client.remove_handler(h, group=group)
                except ValueError:
                    pass

    async def start(self):
        """签到器的入口函数."""

        self.ctx.start(RunStatus.INITIALIZING)

        if self.init_first:
            if not await self.init():
                self.log.warning(f"初始化错误.")
                return self.ctx.finish(RunStatus.FAIL, "初始化错误")

        if (not self.chat_name) and (not self.bot_username):
            raise ValueError("未指定 chat_name 或 bot_username")
        ident = self.chat_name or self.bot_username

        while True:
            try:
                chat = await self.client.get_chat(ident)
            except (UsernameNotOccupied, UsernameInvalid, ChannelInvalid, ChannelPrivate) as e:
                self.log.warning(f'初始化错误: 会话 "{ident}" 已不存在.')
                return self.ctx.finish(RunStatus.IGNORE, "会话已不存在")
            except KeyError as e:
                self.log.info(f"初始化错误: 无法访问, 您可能已被封禁: {e}.")
                show_exception(e)
                return self.ctx.finish(RunStatus.FAIL, "无法访问会话")
            except FloodWait as e:
                if e.value < 360:
                    self.log.info(f"初始化信息: Telegram 要求等待 {e.value} 秒.")
                    await asyncio.sleep(e.value)
                else:
                    self.log.info(
                        f"初始化信息: Telegram 要求等待 {e.value} 秒, 您可能操作过于频繁, 签到器将停止."
                    )
                    return self.ctx.finish(RunStatus.FAIL, "操作过于频繁")
            else:
                break

        is_archived = chat.folder_id == 1

        if await self.client.get_chat_history_count(chat.id) == 0:
            if not self.bot_allow_from_scratch:
                self.log.debug(f'跳过签到: 从未与 "{ident}" 交流.')
                return self.ctx.finish(RunStatus.IGNORE, "从未与该会话交流")

        while True:
            if self.additional_auth:
                for a in self.additional_auth:
                    if not await Link(self.client).auth(a, log_func=self.log.info):
                        return self.ctx.finish(RunStatus.IGNORE, "需要额外认证")

            if not self.init_first:
                if not await self.init():
                    self.log.warning(f"初始化错误.")
                    return self.ctx.finish(RunStatus.FAIL, "初始化错误")

            specs = []
            if self.bot_username:
                try:
                    bot = await self.client.get_users(self.bot_username)
                except IndexError:
                    self.log.warning(f"初始化错误: 用户名 {self.bot_username} 不存在或不是有效用户.")
                    return self.ctx.finish(RunStatus.FAIL, "初始化错误")
                specs.append(f"[green]{bot.name}[/] [gray50](@{bot.username})[/]")
            if chat.title:
                specs.append(f"[green]{chat.title}[/] [gray50](@{chat.username})[/]")
            self.log.info(f"开始执行签到: {' @ '.join(specs)}.")

            if not self.chat_name:
                self.log.debug(f"[gray50]禁用提醒 {self.timeout} 秒: {bot.username}[/]")
                peer = InputNotifyPeer(peer=await self.client.resolve_peer(ident))
                settings: PeerNotifySettings = await self.client.invoke(GetNotifySettings(peer=peer))
                old_mute_until = settings.mute_until
                try:
                    await self.client.mute_chat(ident, time.time() + self.timeout + 10)
                except FloodWait:
                    self.log.debug(f"[gray50]设置禁用提醒因访问超限而失败: {bot.username}[/]")

            self.ctx.status = RunStatus.RUNNING

            try:
                async with self.listener():
                    try:
                        if self.bot_use_history is None:
                            await self.send_checkin()
                        elif not await self.walk_history(self.bot_use_history):
                            await self.send_checkin()
                        await asyncio.wait_for(self.finished.wait(), self.timeout)
                    finally:
                        try:
                            if await asyncio.wait_for(self.cleanup(), 3):
                                self.log.debug(f"[gray50]执行清理成功: {ident}[/]")
                            else:
                                self.log.debug(f"[gray50]执行清理失败: {ident}[/]")
                        except asyncio.TimeoutError:
                            self.log.debug(f"[gray50]执行清理失败: {ident}[/]")
                        if is_archived:
                            self.log.debug(f"[gray50]将会话重新归档: {ident}[/]")
                            try:
                                if await asyncio.wait_for(chat.archive(), 3):
                                    self.log.debug(f"[gray50]重新归档成功: {ident}[/]")
                            except asyncio.TimeoutError:
                                self.log.debug(f"[gray50]归档失败: {ident}[/]")
                        if not self.chat_name:
                            self.log.debug(f"[gray50]将会话设为已读: {ident}[/]")
                            try:
                                if await asyncio.wait_for(self.client.read_chat_history(ident), 3):
                                    self.log.debug(f"[gray50]设为已读成功: {ident}[/]")
                            except asyncio.TimeoutError:
                                self.log.debug(f"[gray50]设为已读失败: {ident}[/]")
            except OSError as e:
                self.log.warning(f'初始化错误: "{e}", 签到器将停止.')
                show_exception(e)
                return self.ctx.finish(RunStatus.FAIL, f"初始化错误: {e}")
            except asyncio.TimeoutError:
                pass
            finally:
                if not self.chat_name:
                    if old_mute_until:
                        try:
                            await asyncio.wait_for(self.client.mute_chat(ident, until=old_mute_until), 3)
                        except asyncio.TimeoutError:
                            self.log.debug(f"[gray50]重新设置通知设置失败: {ident}[/]")
                        except FloodWait:
                            self.log.debug(f"[gray50]重新设置通知设置因访问超限而失败: {ident}[/]")
                        else:
                            self.log.debug(f"[gray50]重新设置通知设置成功: {ident}[/]")
            if not self.finished.is_set():
                self.log.warning("无法在时限内完成签到.")
                return self.ctx.finish(RunStatus.FAIL, "无法在时限内完成签到")
            elif self.current_retries <= self.valid_retries:
                if (
                    self.ctx.status == RunStatus.NONEED
                ):  # 已签到的情况, 如果设置了 checked_retries, 则返回重计划请求
                    if self.checked_retries:
                        if self.ctx.reschedule and self.ctx.reschedule > self.checked_retries:
                            return self.ctx.finish(RunStatus.NONEED)
                        else:
                            now = datetime.now()
                            midnight = now.replace(hour=0, minute=0, second=0, microsecond=0) + timedelta(
                                days=1
                            )
                            max_sleep = midnight - now
                            sleep = timedelta(hours=min((self.ctx.reschedule or 0 + 1) * 1, 6))
                            if sleep > max_sleep:
                                return self.ctx.finish(RunStatus.NONEED)
                            else:
                                self.log.info(f"今日已签到, 即将在 {format_timedelta_human(sleep)} 后重试.")
                                self.ctx.next_time = now + sleep
                                return self.ctx.finish(RunStatus.RESCHEDULE, "等待重新尝试签到")
                return self.ctx.finish()
            else:
                return self.ctx.finish(RunStatus.FAIL, "重试次数超限")

    async def init(self):
        """可重写的初始化函数, 在读取聊天后运行, 在执行签到前运行, 返回 False 将视为初始化错误."""
        return True

    async def cleanup(self):
        """可重写的签到结束后的清理函数, 执行签到成功或失败后运行, 返回 False 将视为清理错误."""
        return True

    async def walk_history(self, limit=0):
        """处理 limit 条历史消息, 并检测是否有验证码."""
        try:
            async for m in self.client.get_chat_history(self.chat_name or self.bot_username, limit=limit):
                if MessageType.CAPTCHA in self.message_type(m):
                    await self.on_photo(m)
                    return True
            return False
        except Exception as e:
            self.log.warning("读取历史消息失败, 将不再读取历史消息.")
            show_exception(e)
            return False

    async def send(self, cmd):
        """向机器人发送命令."""
        if self.chat_name:
            await self.client.send_message(self.chat_name, cmd)
        else:
            await self.client.send_message(self.bot_username, cmd)

    async def send_checkin(self, retry=False):
        """发送签到命令, 或依次发送签到命令序列."""
        cmds = to_iterable(self.bot_checkin_cmd)
        for i, cmd in enumerate(cmds):
            if retry and not i:
                await asyncio.sleep(self.bot_retry_wait)
            if i < len(cmds):
                await asyncio.sleep(self.bot_send_interval)
            await self.send(cmd)

    async def _message_handler(self, client: Client, message: Message):
        """消息处理入口函数的错误处理外壳."""
        try:
            # 创建新的任务并存储
            task = asyncio.create_task(self.message_handler(client, message))
            self._handler_tasks.add(task)
            await task
        except OSError as e:
            self.log.info(f'发生错误: "{e}", 正在重试.')
            show_exception(e)
            await self.retry()
            message.continue_propagation()
        except asyncio.CancelledError:
            task.cancel()
            try:
                await asyncio.wait_for(task, 5)
            except (asyncio.CancelledError, asyncio.TimeoutError):
                pass
            finally:
                raise
        except Exception as e:
            if not config.nofail:
                await self.fail()
                raise
            else:
                self.log.warning(f"发生错误, 签到器将停止.")
                show_exception(e, regular=False)
                await self.fail()
                message.continue_propagation()
        else:
            message.continue_propagation()
        finally:
            if task in self._handler_tasks:
                self._handler_tasks.remove(task)

    async def message_handler(self, client: Client, message: Message, type=None):
        """消息处理入口函数."""
        text = message.text or message.caption
        if text:
            for p, k in self._waiting.items():
                if re.search(p, text):
                    k.set()
                    self._waiting[p] = message
        type = type or self.message_type(message)
        if type:
            if MessageType.TEXT in type:
                await self.on_text(message, message.text)
            if MessageType.CAPTION in type:
                await self.on_text(message, message.caption)
            if MessageType.CAPTCHA in type:
                await self.on_photo(message)

    def message_type(self, message: Message):
        """分析传入消息的类型为验证码或文字."""
        if message.photo:
            if message.caption:
                if self.bot_use_captcha:
                    if self.bot_checkin_caption_pat:
                        if re.search(self.bot_checkin_caption_pat, message.caption):
                            return MessageType.CAPTCHA
                        else:
                            return MessageType.CAPTION
                    else:
                        return MessageType.CAPTCHA
                else:
                    return MessageType.CAPTION
            else:
                if self.bot_use_captcha:
                    return MessageType.CAPTCHA
                else:
                    return MessageType.IGNORE
        elif message.text:
            return MessageType.TEXT
        else:
            return MessageType.IGNORE

    async def on_photo(self, message: Message):
        """分析传入的验证码图片并返回验证码."""
        data = await self.client.download_media(message, in_memory=True)
        ocr = await OCRService.get(
            ocr_name=self.ocr,
            char_range=self.bot_captcha_char_range,
        )

        try:
            with ocr:
                is_gif = getattr(data, "name", "").endswith(".gif")
                ocr_text = await ocr.run(data, gif=is_gif)
                if not ocr_text:
                    self.log.info(f"签到失败: 接收到空验证码, 正在重试.")
                    await self.retry()
                    return

                captcha = ocr_text.translate(str.maketrans("", "", string.punctuation)).replace(" ", "")

            if captcha:
                self.log.debug(f"[gray50]接收验证码: {captcha}.[/]")
                if self.bot_captcha_len and len(captcha) not in to_iterable(self.bot_captcha_len):
                    self.log.info(f"签到失败: 验证码低于设定长度, 正在重试.")
                    await self.retry()
                else:
                    await asyncio.sleep(random.uniform(2, 4))
                    await self.on_captcha(message, captcha)
            else:
                self.log.info(f"签到失败: 接收到空验证码, 正在重试.")
                await self.retry()
        except asyncio.TimeoutError:
            self.log.info("签到失败: 验证码识别失败, 正在重试.")
            await self.retry()
            return

    async def on_captcha(self, message: Message, captcha: str):
        """
        可修改的回调函数.
            message: 包含验证码图片的消息
            captcha: OCR 识别的验证码
        """
        await message.reply(captcha)

    async def on_text(self, message: Message, text: str):
        """接收非验证码消息时, 检测关键词并确认签到成功或失败, 发送用户提示."""
        if not text:
            return
        if any(s in text for s in to_iterable(self.bot_text_ignore)):
            pass
        elif any(
            s in text for s in to_iterable(self.bot_account_fail_keywords) or default_keywords["account_fail"]
        ):
            self.log.warning(f"签到失败: 账户错误.")
            await self.fail()
        elif any(
            s in text
            for s in to_iterable(self.bot_too_many_tries_fail_keywords)
            or default_keywords["too_many_tries_fail"]
        ):
            self.log.warning(f"签到失败: 尝试次数过多.")
            await self.fail()
        elif any(s in text for s in to_iterable(self.bot_checked_keywords) or default_keywords["checked"]):
            self.log.info(f"今日已经签到过了.")
            await self.finish(RunStatus.NONEED, "今日已签到")
        elif any(s in text for s in to_iterable(self.bot_fail_keywords) or default_keywords["fail"]):
            self.log.info(f"签到失败: 验证码错误或网络错误, 正在重试.")
            await self.retry()
        elif any(s in text for s in to_iterable(self.bot_success_keywords) or default_keywords["success"]):
            if await self.before_success():
                if self.bot_success_pat:
                    matches = re.search(self.bot_success_pat, text)
                    if matches:
                        try:
                            self.log.info(
                                f"[yellow]签到成功[/]: + {matches.group(1)} 分 -> {matches.group(2)} 分."
                            )
                        except IndexError:
                            try:
                                self.log.info(f"[yellow]签到成功[/]: 当前/增加 {matches.group(1)} 分.")
                            except IndexError:
                                self.log.info(f"[yellow]签到成功[/].")
                    else:
                        matches = re.search(r"\d+", text)
                        if matches:
                            self.log.info(f"[yellow]签到成功[/]: 当前/增加 {matches.group(0)} 分.")
                        else:
                            self.log.info(f"[yellow]签到成功[/].")
                else:
                    self.log.info(f"[yellow]签到成功[/].")
                await self.after_success()
                await self.finish(RunStatus.SUCCESS, "签到成功")
        else:
            await self.on_unexpected_text(message)

    async def on_unexpected_text(self, message: Message):
        content = message.text or message.caption
        if content:
            spec = content.replace("\n", " ")
            self.log.warning(f"接收到异常返回信息: {spec}, 正在尝试智能回答.")
            if message.reply_markup and message.reply_markup.inline_keyboard:
                buttons = [b.text for r in message.reply_markup.inline_keyboard for b in r]
            else:
                buttons = []
            button_specs = [f"'{b}'" for b in buttons]
            prompt = (
                "我正在进行签到, 机器将显示指令或状态, 我需要通过回答问题以避免被识别为机器人或自动化程序, 并希望完成签到, 现在机器给出的值为:\n\n"
                f"{content}\n\n"
            )
            if buttons:
                prompt += (
                    f"你可选: {', '.join(button_specs)} 中的一个作为回答.\n"
                    "形式为: [CLICK]^XXX^, 其中XXX为回答\n"
                )
            prompt += (
                "如果您认为不应该进行任何操作, 请输出 [NO_RESP], 禁止输出其他内容.\n"
                "如果这是一个指令, 请输出您需要发送或点击的内容.\n"
                "形式为: [SEND]^XXX^, 其中XXX为内容\n"
                "不要说明这是一个指令, 不要说明需要发送文本消息, 仅仅按上述形式输出.\n"
                "如果这是一个状态, 请输出 [IS_STATUS], 禁止输出其他内容."
            )
            for _ in range(3):
                answer, by = await Link(self.client).gpt(prompt)
                if answer:
                    self.log.debug(f"智能回答 ({by}): {answer}")
                    if "[NO_RESP]" in answer:
                        self.log.info(f"智能回答认为无需进行操作, 为了避免风险签到器将停止.")
                        await self.fail()
                        return
                    elif "[IS_STATUS]" in answer:
                        self.log.info(
                            f"智能回答认为这是一条状态信息, 无需进行操作, 为了避免风险签到器将停止."
                        )
                        await self.fail()
                        return
                    elif buttons and "[CLICK]" in answer:
                        self.log.debug(f"当前按钮: {', '.join(button_specs)}")
                        answer_content = re.search(r"\[CLICK\]\^(.+?)\^", answer)
                        if not answer_content:
                            self.log.warning(f"智能回答失败, 为了避免风险签到器将停止.")
                            await self.fail()
                            return
                        answer_content = answer_content.group(1)
                        b, s = process.extractOne(answer_content, buttons, scorer=fuzz.partial_ratio)
                        if s < 70:
                            self.log.info(f"找不到对应回答的按钮, 正在重试.")
                            await asyncio.sleep(3)
                            continue
                        else:
                            try:
                                await message.click(b)
                            except (TimeoutError, MessageIdInvalid):
                                pass
                            self.log.warning(f'智能回答点击了按钮 "{b}", 为了避免风险签到器将停止.')
                            await self.fail()
                            return
                    elif "[SEND]" in answer:
                        answer_content = re.search(r"\[SEND\]\^(.+?)\^", answer)
                        if not answer_content:
                            self.log.warning(f"智能回答失败, 为了避免风险签到器将停止.")
                            await self.fail()
                            return
                        answer_content = answer_content.group(1)
                        await message.reply(answer_content)
                        self.log.warning(f'智能回答回复了 "{answer_content}", 为了避免风险签到器将停止.')
                        await self.fail()
                        return
            else:
                self.log.warning(f"智能回答失败, 为了避免风险签到器将停止.")
                await self.fail()
                return

    async def retry(self):
        """执行重试, 重新发送签到指令."""
        self.current_retries += 1
        if self.current_retries <= self.valid_retries:
            await asyncio.sleep(self.bot_retry_wait)
            await self.send_checkin(retry=True)
        else:
            self.log.warning("超过最大重试次数.")
            await self.finish(RunStatus.FAIL, "超过最大重试次数")

    async def fail(self, status: RunStatus = RunStatus.FAIL, message: str = None):
        """设定签到器失败."""
        return await self.finish(status, message, fail=True)

    async def finish(self, status: RunStatus = RunStatus.SUCCESS, message: str = None, fail=False):
        """设定签到器结束."""

        if fail:
            self.current_retries = float("inf")
        self.ctx.status = status
        self.ctx.status_info = message
        self.finished.set()
        return self

    async def wait_until(self, pattern: str = ".", timeout: float = None):
        """等待特定消息出现."""
        self._waiting[pattern] = e = asyncio.Event()
        try:
            await asyncio.wait_for(e.wait(), timeout)
        except asyncio.TimeoutError:
            return None
        else:
            msg: Message = self._waiting[pattern]
            return msg

    async def before_success(self):
        """签到成功前钩子, 返回值为 True 将继续执行并显示成功信息."""
        return True

    async def after_success(self):
        """签到成功后钩子."""
        pass


class AnswerBotCheckin(BotCheckin):
    """签到类, 用于按钮模式签到."""

    bot_answer_button_message_pat: str = None  # 回答按键消息内容的 regex 条件
    bot_answer_button_pat: str = None  # 所有按键需要满足的 regex 条件

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.mutex = asyncio.Lock()  # 实例变量异步锁
        self.operable = asyncio.Condition(self.mutex)  # 当 message 被设定时提醒等待的异步线程.
        self.message: Message = None  # 存储可用于回复的带按钮的消息.

    async def walk_history(self, limit=0):
        """处理 limit 条历史消息, 并检测是否有验证码或按钮."""
        try:
            answer = None
            captcha = None
            async for m in self.client.get_chat_history(self.chat_name or self.bot_username, limit=limit):
                if MessageType.ANSWER in self.message_type(m):
                    answer = answer or m
                if MessageType.CAPTCHA in self.message_type(m):
                    captcha = captcha or m
                if answer and captcha:
                    break
            else:
                return False
            await self.on_answer(answer)
            await self.on_photo(captcha)
            return True
        except Exception as e:
            self.log.warning("读取历史消息失败, 将不再读取历史消息.")
            show_exception(e)
            return False

    def get_keys(self, message: Message):
        """获得所有按钮信息."""
        reply_markup = message.reply_markup
        if isinstance(reply_markup, InlineKeyboardMarkup):
            return [k.text for r in reply_markup.inline_keyboard for k in r]
        elif isinstance(reply_markup, ReplyKeyboardMarkup):
            return [k.text for r in reply_markup.keyboard for k in r]

    def is_valid_answer(self, message: Message):
        """确认消息是回复按钮消息."""
        if not message.reply_markup:
            return False
        if self.bot_answer_button_message_pat:
            text = message.text or message.caption
            if not text:
                return False
            if not re.search(self.bot_answer_button_message_pat, text):
                return False
        if self.bot_answer_button_pat:
            for k in self.get_keys(message):
                if not re.search(self.bot_answer_button_pat, k):
                    return False
        return True

    def message_type(self, message: Message):
        """分析传入消息的类型为验证码或文字."""
        if self.is_valid_answer(message):
            return MessageType.ANSWER | super().message_type(message)
        else:
            return super().message_type(message)

    async def message_handler(self, client: Client, message: Message, type=None):
        """消息处理入口函数."""
        type = type or self.message_type(message)
        if MessageType.ANSWER in type:
            await self.on_answer(message)
        await super().message_handler(client, message, type=type)

    async def on_answer(self, message: Message):
        """当检测到带按钮的用于回答的消息时保存其信息."""
        async with self.mutex:
            if self.message:
                if self.message.date > message.date:
                    return
                else:
                    self.message = message
            else:
                self.message = message
                self.operable.notify()

    async def on_captcha(self, message: Message, captcha: str):
        """
        可修改的回调函数.
        参数:
            message: 包含验证码图片的消息
            captcha: OCR 识别的验证码
        """
        async with self.operable:
            if not self.message:
                await self.operable.wait()
            # Convert captcha to lowercase and compare with lowercase keys
            match = [(k, fuzz.ratio(k.lower(), captcha.lower())) for k in self.get_keys(self.message)]
            max_k, max_r = max(match, key=lambda x: x[1])
            if max_r < 75:
                self.log.info(f'未能找到对应 "{captcha}" 的按键, 正在重试.')
                await self.retry()
            else:
                try:
                    await self.message.click(max_k)
                except (TimeoutError, MessageIdInvalid, DataInvalid):
                    pass



from __future__ import annotations

import base64
from collections import OrderedDict
from contextlib import asynccontextmanager
from datetime import datetime
import asyncio
import inspect
import os
from pathlib import Path
import sqlite3
import struct
from typing import Union
import logging

from rich.prompt import Prompt
from loguru import logger
import pyrogram
from pyrogram import raw, types, utils, filters, dispatcher
from pyrogram.enums import SentCodeType
from pyrogram.errors import (
    ChannelPrivate,
    PersistentTimestampOutdated,
    PersistentTimestampInvalid,
    BadRequest,
    SessionPasswordNeeded,
    CodeInvalid,
    PhoneCodeInvalid,
    FloodWait,
    PhoneNumberInvalid,
    PhoneNumberBanned,
    BadRequest,
    MessageIdInvalid,
)
from pyrogram.handlers import (
    MessageHandler,
    RawUpdateHandler,
    DisconnectHandler,
    EditedMessageHandler,
)
from pyrogram.storage.memory_storage import MemoryStorage
from pyrogram.storage.sqlite_storage import SQLiteStorage
from pyrogram.storage.file_storage import USERNAMES_SCHEMA, UPDATE_STATE_SCHEMA
from pyrogram.handlers.handler import Handler

from embykeeper import var, __name__ as __product__, __version__
from embykeeper.utils import async_partial, show_exception

var.tele_used.set()

logger = logger.bind(scheme="telegram")


def _name(self: Union[types.User, types.Chat]):
    return " ".join([n for n in (self.first_name, self.last_name) if n])


def _chat_name(self: types.Chat):
    if self.title:
        return self.title
    else:
        return _name(self)


setattr(types.User, "name", property(_name))
setattr(types.Chat, "name", property(_chat_name))


class LogRedirector(logging.StreamHandler):
    def emit(self, record):
        try:
            if record.levelno >= logging.WARNING:
                logger.debug(f"Pyrogram log: {record.getMessage()}")
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


pyrogram_session_logger = logging.getLogger("pyrogram")
for h in pyrogram_session_logger.handlers[:]:
    pyrogram_session_logger.removeHandler(h)
pyrogram_session_logger.addHandler(LogRedirector())


class Dispatcher(dispatcher.Dispatcher):
    updates_count = 0

    def __init__(self, client: Client):
        super().__init__(client)
        self.mutex = asyncio.Lock()

    async def start(self):
        logger.debug("Telegram 更新分配器启动.")
        if not self.client.no_updates:

            self.handler_worker_tasks = []
            for _ in range(self.client.workers):
                self.handler_worker_tasks.append(self.loop.create_task(self.handler_worker()))

            if not self.client.skip_updates:
                await self.client.recover_gaps()

    async def stop(self, clear: bool = True):
        if not self.client.no_updates:
            for i in range(self.client.workers):
                self.updates_queue.put_nowait(None)
            for i in self.handler_worker_tasks:
                i.cancel()
                try:
                    await i
                except asyncio.CancelledError:
                    pass
            if clear:
                self.handler_worker_tasks.clear()
                self.groups.clear()

    def add_handler(self, handler, group: int):
        async def fn():
            async with self.mutex:
                if group not in self.groups:
                    self.groups[group] = []
                    self.groups = OrderedDict(sorted(self.groups.items()))
                self.groups[group].append(handler)
                # logger.debug(f"增加了 Telegram 更新处理器: {handler.__class__.__name__}.")

        return self.loop.create_task(fn())

    def remove_handler(self, handler, group: int):
        async def fn():
            async with self.mutex:
                if group not in self.groups:
                    raise ValueError(f"Group {group} does not exist. Handler was not removed.")
                self.groups[group].remove(handler)
                # logger.debug(f"移除了 Telegram 更新处理器: {handler.__class__.__name__}.")

        return self.loop.create_task(fn())

    async def handler_worker(self):
        while True:
            packet = await self.updates_queue.get()
            Dispatcher.updates_count += 1

            if packet is None:
                break

            try:
                update, users, chats = packet
                parser = self.update_parsers.get(type(update), None)

                try:
                    parsed_update, handler_type = (
                        await parser(update, users, chats) if parser is not None else (None, type(None))
                    )
                except (ValueError, BadRequest):
                    continue

                async with self.mutex:
                    groups = {i: g[:] for i, g in self.groups.items()}

                for group in groups.values():
                    for handler in group:
                        args = None

                        if isinstance(handler, handler_type):
                            try:
                                if await handler.check(self.client, parsed_update):
                                    args = (parsed_update,)
                            except Exception as e:
                                logger.warning(f"Telegram 错误: {e}")
                                continue

                        elif isinstance(handler, RawUpdateHandler):
                            try:
                                if await handler.check(self.client, update):
                                    args = (update, users, chats)
                            except Exception as e:
                                logger.debug(f"更新回调函数内发生错误.")
                                show_exception(e, regular=False)
                        if args is None:
                            continue

                        try:
                            if inspect.iscoroutinefunction(handler.callback):
                                await handler.callback(self.client, *args)
                            else:
                                await self.loop.run_in_executor(
                                    self.client.executor, handler.callback, self.client, *args
                                )
                        except pyrogram.StopPropagation:
                            raise
                        except pyrogram.ContinuePropagation:
                            continue
                        except Exception as e:
                            logger.error(f"更新回调函数内发生错误.")
                            show_exception(e, regular=False)
                        break
                    else:
                        continue
                    break
            except pyrogram.StopPropagation:
                pass
            except Exception as e:
                logger.warning("更新控制器错误.")
                show_exception(e, regular=False)


class FileStorage(SQLiteStorage):
    FILE_EXTENSION = ".session"

    def __init__(self, name: str, workdir: Path, session_string: str = None):
        super().__init__(name)

        self.database = workdir / (self.name + self.FILE_EXTENSION)
        self.session_string = session_string

    def update(self):
        version = self.version()

        if version == 1:
            with self.conn:
                self.conn.execute("DELETE FROM peers")

            version += 1

        if version == 2:
            with self.conn:
                self.conn.execute("ALTER TABLE sessions ADD api_id INTEGER")

            version += 1

        if version == 3:
            with self.conn:
                self.conn.executescript(USERNAMES_SCHEMA)

            version += 1

        if version == 4:
            with self.conn:
                self.conn.executescript(UPDATE_STATE_SCHEMA)

            version += 1

        if version == 5:
            with self.conn:
                self.conn.execute("CREATE INDEX idx_usernames_id ON usernames (id);")

            version += 1

        self.version(version)

    async def open(self):
        path = self.database
        file_exists = path.is_file()

        self.conn = sqlite3.connect(str(path), timeout=1, check_same_thread=False)

        if not file_exists:
            self.create()
            if self.session_string:
                # Old format
                if len(self.session_string) in [self.SESSION_STRING_SIZE, self.SESSION_STRING_SIZE_64]:
                    dc_id, test_mode, auth_key, user_id, is_bot = struct.unpack(
                        (
                            self.OLD_SESSION_STRING_FORMAT
                            if len(self.session_string) == self.SESSION_STRING_SIZE
                            else self.OLD_SESSION_STRING_FORMAT_64
                        ),
                        base64.urlsafe_b64decode(self.session_string + "=" * (-len(self.session_string) % 4)),
                    )

                    await self.dc_id(dc_id)
                    await self.test_mode(test_mode)
                    await self.auth_key(auth_key)
                    await self.user_id(user_id)
                    await self.is_bot(is_bot)
                    await self.date(0)

                    logger.warning(
                        "You are using an old session string format. Use export_session_string to update"
                    )
                    return

                dc_id, api_id, test_mode, auth_key, user_id, is_bot = struct.unpack(
                    self.SESSION_STRING_FORMAT,
                    base64.urlsafe_b64decode(self.session_string + "=" * (-len(self.session_string) % 4)),
                )

                await self.dc_id(dc_id)
                await self.api_id(api_id)
                await self.test_mode(test_mode)
                await self.auth_key(auth_key)
                await self.user_id(user_id)
                await self.is_bot(is_bot)
                await self.date(0)
        else:
            self.update()

        with self.conn:
            self.conn.execute("VACUUM")

    async def delete(self):
        os.remove(self.database)


class Client(pyrogram.Client):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.dispatcher = Dispatcher(self)
        self.stop_handlers = []
        if self.in_memory:
            self.storage = MemoryStorage(self.name, self.session_string)
        else:
            self.storage = FileStorage(self.name, self.workdir, self.session_string)

    async def authorize(self):
        if self.bot_token:
            return await self.sign_in_bot(self.bot_token)
        retry = False
        while True:
            try:
                sent_code = await self.send_code(self.phone_number)
                code_target = {
                    SentCodeType.APP: " Telegram 客户端",
                    SentCodeType.SMS: "短信",
                    SentCodeType.CALL: "来电",
                    SentCodeType.FLASH_CALL: "闪存呼叫",
                    SentCodeType.FRAGMENT_SMS: " Fragment 短信",
                    SentCodeType.EMAIL_CODE: "邮件",
                }
                if not self.phone_code:
                    if retry:
                        msg = f'验证码错误, 请重新输入 "{self.phone_number}" 的登录验证码 (按回车确认)'
                    else:
                        msg = f'请从{code_target[sent_code.type]}接收 "{self.phone_number}" 的登录验证码 (按回车确认)'
                    try:
                        self.phone_code = Prompt.ask(" " * 23 + msg, console=var.console)
                    except EOFError:
                        raise BadRequest(
                            f'登录 "{self.phone_number}" 时出现异常: 您正在使用非交互式终端, 无法输入验证码.'
                        )
                signed_in = await self.sign_in(self.phone_number, sent_code.phone_code_hash, self.phone_code)
            except (CodeInvalid, PhoneCodeInvalid):
                self.phone_code = None
                retry = True
            except SessionPasswordNeeded:
                retry = False
                while True:
                    if not self.password:
                        if retry:
                            msg = f'密码错误, 请重新输入 "{self.phone_number}" 的两步验证密码 (不显示, 按回车确认)'
                        else:
                            msg = f'需要输入 "{self.phone_number}" 的两步验证密码 (不显示, 按回车确认)'
                        self.password = Prompt.ask(" " * 23 + msg, password=True, console=var.console)
                    try:
                        return await self.check_password(self.password)
                    except BadRequest:
                        self.password = None
                        retry = True
            except FloodWait:
                raise BadRequest(f'登录 "{self.phone_number}" 时出现异常: 登录过于频繁.')
            except PhoneNumberInvalid:
                raise BadRequest(
                    f'登录 "{self.phone_number}" 时出现异常: 您使用了错误的手机号 (格式错误或没有注册).'
                )
            except PhoneNumberBanned:
                raise BadRequest(f'登录 "{self.phone_number}" 时出现异常: 您的账户已被封禁.')
            except Exception as e:
                logger.error(f"登录时出现异常错误!")
                show_exception(e, regular=False)
                retry = True
            else:
                break
        if isinstance(signed_in, types.User):
            return signed_in
        else:
            raise BadRequest("该账户尚未注册")

    def add_handler(self, handler: Handler, group: int = 0):
        if isinstance(handler, DisconnectHandler):
            self.disconnect_handler = handler.callback

            async def dummy():
                pass

            return asyncio.ensure_future(dummy())
        else:
            return self.dispatcher.add_handler(handler, group)

    def remove_handler(self, handler: Handler, group: int = 0):
        if isinstance(handler, DisconnectHandler):
            self.disconnect_handler = None

            async def dummy():
                pass

            return asyncio.ensure_future(dummy())
        else:
            return self.dispatcher.remove_handler(handler, group)

    async def invoke(self, query, *args, **kw):
        last_error = None
        for _ in range(3):
            try:
                return await super().invoke(query, *args, **kw)
            except OSError as e:
                last_error = e
                await asyncio.sleep(0.5)
                continue
            except (
                sqlite3.ProgrammingError
            ) as e:  # Cannot operate on a closed database. The client is stopping.
                raise asyncio.CancelledError() from None
        else:
            raise OSError(
                f"执行 Telegram 请求由于网络因素而失败, 且重试超限."
                f"({last_error.__class__.__name__} for {query.__class__.__name__})"
            )

    @asynccontextmanager
    async def catch_reply(self, chat_id: Union[int, str], outgoing=False, filter=None):
        async def handler_func(client, message, future: asyncio.Future):
            try:
                future.set_result(message)
            except asyncio.InvalidStateError:
                pass

        future = asyncio.Future()
        f = filters.chat(chat_id)
        if not outgoing:
            f = f & (~filters.outgoing)
        if filter:
            f = f & filter
        handler = MessageHandler(async_partial(handler_func, future=future), f)
        await self.add_handler(handler, group=0)
        try:
            yield future
        finally:
            await self.remove_handler(handler, group=0)

    @asynccontextmanager
    async def catch_edit(self, message: types.Message, filter=None):
        def filter_message(id: int):
            async def func(flt, _, message: types.Message):
                return message.id == id

            return filters.create(func, "MessageFilter")

        async def handler_func(client, message, future: asyncio.Future):
            try:
                future.set_result(message)
            except asyncio.InvalidStateError:
                pass

        future = asyncio.Future()
        f = filter_message(message.id)
        if filter:
            f = f & filter
        handler = EditedMessageHandler(async_partial(handler_func, future=future), f)
        await self.add_handler(handler, group=0)
        try:
            yield future
        finally:
            await self.remove_handler(handler, group=0)

    async def wait_reply(
        self,
        chat_id: Union[int, str],
        send: str = None,
        timeout: float = 10,
        outgoing=False,
        filter=None,
    ):
        async with self.catch_reply(chat_id=chat_id, filter=filter) as f:
            if send:
                await self.send_message(chat_id, send)
            msg: types.Message = await asyncio.wait_for(f, timeout)
            return msg

    async def wait_edit(
        self,
        message: types.Message,
        click: Union[str, int] = None,
        timeout: float = 10,
        noanswer=True,
        filter=None,
    ):
        async with self.catch_edit(message, filter=filter) as f:
            if click:
                try:
                    await message.click(click)
                except (TimeoutError, MessageIdInvalid):
                    if noanswer:
                        pass
                    else:
                        raise
            msg: types.Message = await asyncio.wait_for(f, timeout)
            return msg

    async def mute_chat(self, chat_id: Union[int, str], until: Union[int, datetime, None] = None):
        if until is None:
            until = 0x7FFFFFFF  # permanent mute
        elif isinstance(until, datetime):
            until = until.timestamp()
        return await self.invoke(
            raw.functions.account.UpdateNotifySettings(
                peer=raw.types.InputNotifyPeer(peer=await self.resolve_peer(chat_id)),
                settings=raw.types.InputPeerNotifySettings(
                    show_previews=False,
                    mute_until=int(until),
                ),
            )
        )

    async def handle_updates(self, updates):
        if isinstance(updates, (raw.types.Updates, raw.types.UpdatesCombined)):
            is_min = any(
                (
                    await self.fetch_peers(updates.users),
                    await self.fetch_peers(updates.chats),
                )
            )

            users = {u.id: u for u in updates.users}
            chats = {c.id: c for c in updates.chats}

            for update in updates.updates:
                channel_id = getattr(
                    getattr(getattr(update, "message", None), "peer_id", None), "channel_id", None
                ) or getattr(update, "channel_id", None)

                pts = getattr(update, "pts", None)
                pts_count = getattr(update, "pts_count", None)

                if pts and not self.skip_updates:
                    await self.storage.update_state(
                        (
                            utils.get_channel_id(channel_id) if channel_id else 0,
                            pts,
                            None,
                            updates.date,
                            updates.seq,
                        )
                    )

                if isinstance(update, raw.types.UpdateNewChannelMessage) and is_min:
                    message = update.message

                    if not isinstance(message, raw.types.MessageEmpty):
                        try:
                            diff = await self.invoke(
                                raw.functions.updates.GetChannelDifference(
                                    channel=await self.resolve_peer(utils.get_channel_id(channel_id)),
                                    filter=raw.types.ChannelMessagesFilter(
                                        ranges=[
                                            raw.types.MessageRange(
                                                min_id=update.message.id, max_id=update.message.id
                                            )
                                        ]
                                    ),
                                    pts=pts - pts_count,
                                    limit=pts,
                                    force=False,
                                )
                            )
                        except (ChannelPrivate, PersistentTimestampOutdated, PersistentTimestampInvalid):
                            pass
                        else:
                            if not isinstance(diff, raw.types.updates.ChannelDifferenceEmpty):
                                users.update({u.id: u for u in diff.users})
                                chats.update({c.id: c for c in diff.chats})

                self.dispatcher.updates_queue.put_nowait((update, users, chats))

        elif isinstance(updates, (raw.types.UpdateShortMessage, raw.types.UpdateShortChatMessage)):
            if not self.skip_updates:
                await self.storage.update_state((0, updates.pts, None, updates.date, None))

            diff = await self.invoke(
                raw.functions.updates.GetDifference(
                    pts=updates.pts - updates.pts_count, date=updates.date, qts=-1
                )
            )

            if diff.new_messages:
                self.dispatcher.updates_queue.put_nowait(
                    (
                        raw.types.UpdateNewMessage(
                            message=diff.new_messages[0],
                            pts=updates.pts,
                            pts_count=updates.pts_count,
                        ),
                        {u.id: u for u in diff.users},
                        {c.id: c for c in diff.chats},
                    )
                )
            else:
                if diff.other_updates:  # The other_updates list can be empty
                    self.dispatcher.updates_queue.put_nowait((diff.other_updates[0], {}, {}))
        elif isinstance(updates, raw.types.UpdateShort):
            self.dispatcher.updates_queue.put_nowait((updates.update, {}, {}))



from __future__ import annotations

import inspect
from logging import Formatter
import asyncio
import logging
from typing import TYPE_CHECKING

from loguru import logger
from rich.logging import RichHandler
from rich.text import Text

from . import var
from .utils import to_iterable

if TYPE_CHECKING:
    from loguru import Logger

pad = " " * 23

scheme_names = {
    "telegram": "Telegram",
    "telechecker": "每日签到",
    "telemonitor": "消息监控",
    "telemessager": "定时水群",
    "telelink": "账号服务",
    "telenotifier": "消息推送",
    "embywatcher": "Emby保活",
    "subsonic": "Subsonic保活",
    "datamanager": "下载器",
    "debugtool": "开发工具",
    "config": "配置文件",
    "cfsolver": "验证解析",
}


def formatter(record):
    """根据日志器的 scheme 属性配置输出格式."""
    extra = record["extra"]
    scheme = extra.get("scheme", None)

    def ifextra(keys, pattern="{}"):
        keys = to_iterable(keys)
        if all(k in extra for k in keys):
            return pattern.format(*[f"{{extra[{k}]}}" for k in keys])
        else:
            return ""

    if scheme in ("telegram", "telechecker", "telemonitor", "telemessager", "telelink"):
        username = ifextra("username", " ([cyan]{}[/])")
        name = ifextra("name", "([magenta]{}[/]) ")
        return f"[blue]{scheme_names[scheme]}[/]{username}: {name}{{message}}"
    elif scheme == "embywatcher":
        ident = ifextra(["username", "server"], " ([cyan]{}@{}[/])")
        return f"[blue]{scheme_names[scheme]}[/]{ident}: {{message}}"
    elif scheme == "subsonic":
        ident = ifextra(["username", "server"], " ([cyan]{}@{}[/])")
        return f"[blue]{scheme_names[scheme]}[/]{ident}: {{message}}"
    elif scheme in ("datamanager", "debugtool", "config", "cfsolver"):
        return f"[blue]{scheme_names[scheme]}[/]: {{message}}"
    else:
        return "{message}"


def initialize(level="INFO", **kw):
    """初始化日志配置."""

    from asyncio import constants

    logger.remove()
    handler = RichHandler(
        console=var.console, markup=True, rich_tracebacks=True, tracebacks_suppress=[asyncio], **kw
    )
    handler.setFormatter(Formatter(None, "[%m/%d %H:%M]"))
    logger.add(handler, format=formatter, level=level, colorize=False)

    constants.LOG_THRESHOLD_FOR_CONNLOST_WRITES = 1000000


class InterceptHandler(logging.Handler):
    """A logging handler that intercepts standard logging messages and redirects them to loguru."""

    def __init__(self, level: int = 0):
        super().__init__(level)

    def emit(self, record):
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno
        frame, depth = inspect.currentframe(), 0
        while frame and (depth == 0 or frame.f_code.co_filename == logging.__file__):
            frame = frame.f_back
            depth += 1
        text = record.getMessage()
        text = Text.from_ansi(text).plain
        text = f"[{level}] ({record.name}) {text}"
        logger.opt(depth=depth, exception=record.exc_info).debug(text)


def apply_logging_adapter(level: int = 20):
    logging.basicConfig(handlers=[InterceptHandler()], level=level, force=True)



import asyncio
from datetime import datetime, time, timedelta
from dateutil import parser
import re
from typing import Callable, Union
import json
import hashlib

from loguru import logger

from .config import config
from .runinfo import RunContext, RunStatus
from .utils import next_random_datetime


class Scheduler:
    """异步函数计划执行器"""

    @classmethod
    def from_str(
        cls,
        func: Callable,
        interval_days: str,
        time_range: str,
        **kw,
    ):
        """从字符串创建调度器

        Args:
            func: 要执行的异步函数
            interval_days: 间隔天数字符串，支持数字或 "<min,max>" 格式
            time_range: 时间范围字符串，支持具体时间或 "<start,end>" 格式
        Returns:
            Scheduler: 调度器实例
        """
        # Parse interval days
        interval_range_match = re.match(r"<\s*(\d+)\s*,\s*(\d+)\s*>", interval_days)
        if interval_range_match:
            days = [int(interval_range_match.group(1)), int(interval_range_match.group(2))]
        else:
            try:
                days = abs(int(interval_days))
            except ValueError:
                raise ValueError(f"无法解析间隔天数: {interval_days}")

        # Parse time range
        time_range_match = re.match(r"<\s*(.*?)\s*,\s*(.*?)\s*>", time_range)
        if time_range_match:
            start_time, end_time = time_range_match.group(1), time_range_match.group(2)
        else:
            start_time = end_time = time_range

        return cls(
            func,
            days=days,
            start_time=start_time,
            end_time=end_time,
            **kw,
        )

    def __init__(
        self,
        func: Callable,
        days: Union[int, list] = 1,
        start_time: Union[str, time] = None,
        end_time: Union[str, time] = None,
        sid: str = None,
        description: str = None,
        on_next_time: Callable[[datetime], None] = None,
    ):
        """
        Args:
            func: 要执行的异步函数
            days: 执行间隔天数，可以是固定天数或者[最小天数, 最大天数]
            start_time: 执行时间范围起始时间 (可选)
            end_time: 执行时间范围结束时间 (可选)
            sid: 调度器ID, 用于缓存下次执行时间
            description: 调度器描述
            on_next_time: 回调函数，在计算出下一次执行时间时调用
        """
        self.func = func
        if config.debug_cron:
            logger.warning(f"计划任务调试模式下任务开始时间被调整为10秒后: {description}")
            debug_time = (datetime.now() + timedelta(seconds=10)).time()
            self.days = 0
            self.start_time = debug_time
            self.end_time = debug_time
        else:
            self.days = days
            self.start_time = self._parse_time(start_time)
            self.end_time = self._parse_time(end_time)
        self.sid = sid
        self.description = description
        self.on_next_time = on_next_time
        self._cache_key = f"scheduler.{sid}" if sid else None
        self._next_time = None
        self._ctx: RunContext = None

    def _parse_time(self, t):
        if isinstance(t, str):
            return parser.parse(t).time()
        return t

    def _get_scheduler_config(self):
        """获取调度器配置的哈希值"""
        config = {
            "days": self.days,
            "start_time": self.start_time.isoformat() if self.start_time else None,
            "end_time": self.end_time.isoformat() if self.end_time else None,
        }
        # Convert config to a stable string representation and hash it
        config_str = json.dumps(config, sort_keys=True)
        return hashlib.md5(config_str.encode()).hexdigest()

    @property
    def next_time(self) -> datetime:
        """获取下一次执行时间"""
        if not self._next_time:
            self._next_time = self._get_next_time()
        return self._next_time

    def _get_next_time(self) -> datetime:
        """计算或获取缓存的下一次执行时间"""
        from .cache import cache

        now = datetime.now()
        next_time = None

        # Try to get cached next execution time
        if self._cache_key:
            cached = cache.get(self._cache_key)
            if cached:
                cached_config_hash = cached.get("config_hash")
                cached_time = cached.get("next_time")

                # Check if config hash matches and time hasn't passed
                if (
                    cached_config_hash == self._get_scheduler_config()
                    and cached_time
                    and parser.parse(cached_time) > now
                ):
                    next_time = parser.parse(cached_time)

        # Calculate new next_time if needed
        if not next_time:
            # Calculate interval days
            if isinstance(self.days, (list, tuple)):
                interval = self.days[0] + (self.days[1] - self.days[0])
            else:
                interval = self.days

            next_time = next_random_datetime(
                start_time=self.start_time, end_time=self.end_time, interval_days=interval
            )

            # Cache the next execution time with config hash
            if self._cache_key:
                cache.set(
                    self._cache_key,
                    {
                        "config_hash": self._get_scheduler_config(),
                        "next_time": next_time.isoformat(),
                        "description": self.description,
                    },
                )

        return next_time

    async def schedule(self):
        """等待到指定时间范围内执行函数"""
        from .cache import cache

        while True:
            now = datetime.now()
            self._next_time = self._get_next_time()

            # Call the hook function if provided
            if self.on_next_time:
                self._ctx = self.on_next_time(self._next_time)

            # Wait until the scheduled time
            wait_seconds = (self._next_time - now).total_seconds()
            if wait_seconds > 0:
                await asyncio.sleep(wait_seconds)

            # Execute the function
            try:
                try:
                    # Shield the function execution to distinguish cancellation source
                    await asyncio.shield(self.func(self._ctx))
                except asyncio.CancelledError:
                    # This is a cancellation from within self.func
                    if self._ctx:
                        self._ctx.finish(RunStatus.ERROR, "任务在运行时被取消")
                    raise  # Re-raise to be caught by outer try block
            except asyncio.CancelledError:
                # This is a cancellation from outside schedule()
                if self._ctx:
                    self._ctx.finish(RunStatus.CANCELLED, "任务被取消")
                raise  # Re-raise to propagate cancellation
            except Exception:
                if self._ctx:
                    self._ctx.finish(RunStatus.ERROR, f"任务发生错误")
                if not config.nofail:
                    raise

            if self._cache_key:
                try:
                    cache.delete(self._cache_key)
                except KeyError:
                    pass
            self._ctx = None
            self._next_time = None

            # If days is 0, break the loop after one execution
            if isinstance(self.days, (list, tuple)) and self.days[0] == 0:
                break



from __future__ import annotations

import asyncio
import re
from typing import TYPE_CHECKING

from pyrogram.errors import MessageIdInvalid
from pyrogram.types import Message
from pyrogram.raw.types.messages import BotCallbackAnswer

from .pyrogram import Client

if TYPE_CHECKING:
    from loguru import Logger


class EmbybossRegister:
    def __init__(self, client: Client, logger: Logger, username: str, password: str):
        self.client = client
        self.log = logger
        self.username = username
        self.password = password

    async def run(self, bot: str):
        try:
            panel = await self.client.wait_reply(bot, "/start")
        except asyncio.TimeoutError:
            self.log.warning("初始命令无响应, 无法注册.")
            return False

        text = panel.text or panel.caption
        try:
            current_status = re.search(r"当前状态 \| ([^\n]+)", text).group(1).strip()
            register_status = re.search(r"注册状态 \| (True|False)", text).group(1) == "True"
            available_slots = int(re.search(r"可注册席位 \| (\d+)", text).group(1))
        except (AttributeError, ValueError):
            self.log.warning("无法解析界面, 无法注册, 可能您已注册.")
            return False

        if current_status != "未注册":
            self.log.warning("当前状态不是未注册, 无法注册.")
            return False

        if not register_status:
            self.log.debug("未开注, 将继续监控.")
            return False

        if available_slots <= 0:
            self.log.debug("可注册席位不足, 将继续监控.")
            return False

        # 点击创建账户按钮
        buttons = panel.reply_markup.inline_keyboard
        create_button = None
        for row in buttons:
            for button in row:
                if "创建账户" in button.text:
                    create_button = button.text
                    break
            if create_button:
                break

        if not create_button:
            self.log.warning("找不到创建账户按钮, 无法注册.")
            return False

        async with self.client.catch_reply(panel.chat.id) as f:
            try:
                answer: BotCallbackAnswer = await panel.click(create_button)
                if "已关闭" in answer.message:
                    self.log.debug("未开注, 将继续监控.")
                    return False
            except (TimeoutError, MessageIdInvalid):
                pass
            try:
                msg: Message = await asyncio.wait_for(f, 5)
            except asyncio.TimeoutError:
                self.log.warning("创建账户按钮点击无响应, 无法注册.")
                return False

        text = msg.text or msg.caption
        if "您已进入注册状态" not in text:
            self.log.warning("未能正常进入注册状态, 注册失败.")
            return False

        try:
            msg = await self.client.wait_reply(msg.chat.id, f"{self.username} {self.password}")
        except asyncio.TimeoutError:
            self.log.warning("发送凭据后无响应, 无法注册.")
            return False

        msg = await self.client.wait_edit(msg)
        text = msg.text or msg.caption
        if "创建用户成功" not in text:
            self.log.warning("发送凭据后注册失败.")
            return False
        else:
            return True



from pyrogram.types import Message

from . import Monitor


class InfinityFlyMonitor(Monitor):
    name = "Infinity Fly"
    chat_keyword = r"register-[-\w]{36}"
    bot_username = "Infinity94bot"
    chat_name = "infinityfly664"
    notify_create_name = True
    allow_edit = False
    additional_auth = ["prime"]

    async def on_trigger(self, message: Message, key, reply):
        await self.client.send_message(self.bot_username, f"/invite {key}")
        await self.client.send_message(self.bot_username, f"/create {self.unique_name}")
        self.log.bind(msg=True).info(f'已向 Bot @{self.bot_username} 发送了邀请码: "{key}", 请查看.')



import random
import string
from loguru import logger

from embykeeper.config import config

logger = logger.bind(scheme="cfsolver")


async def get_cf_clearance(url: str, proxy: str = None):
    from embykeeper.telegram.link import Link
    from embykeeper.telegram.session import ClientsSession
    from embykeeper.wssocks import WSSocks

    telegrams = config.telegram.account
    if not len(telegrams):
        logger.warning(f"未设置 Telegram 账号, 无法使用验证码解析.")
    async with ClientsSession(telegrams[:1]) as clients:
        async for _, tg in clients:
            for i in range(3):
                ws_url, token = await Link(tg).wssocks()
                if not token:
                    logger.warning(f"反向代理服务器申请失败, 正在重试 ({i+1}/3).")
                    continue
                wssocks = WSSocks(proxy=proxy)
                try:
                    connector_token = "".join(random.choices(string.ascii_letters + string.digits, k=16))
                    output = await wssocks.start(ws_url, token, connector_token, proxy)
                    if output:
                        logger.warning(f"连接到反向代理服务器失败, 正在重试 ({i+1}/3)")
                        logger.debug(f"WSSocks 输出:\n{output}")
                        continue
                    cf_clearance, useragent = await Link(tg).captcha_wssocks(connector_token, url)
                    if not cf_clearance:
                        logger.warning(f"解析失败, 正在重试 ({i+1}/3).")
                        continue
                    return cf_clearance, useragent
                finally:
                    wssocks.stop()
            else:
                return None, None



import os
from pathlib import Path
from typer.testing import CliRunner

import pytest

import embykeeper
from embykeeper.cli import app

runner = CliRunner()


@pytest.fixture()
def in_temp_dir(tmp_path: Path):
    current = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(current)


def test_version():
    result = runner.invoke(app, ["--version"])
    assert embykeeper.__version__ in result.stdout
    assert result.exit_code == 0


def test_create_config(in_temp_dir: Path):
    result = runner.invoke(app, ["--example-config"])
    assert "这是一个配置文件范例" in result.stdout
    assert result.exit_code == 0



import asyncio
import io

from rich.text import Text
from loguru import logger

from embykeeper.schema import TelegramAccount
from embykeeper.utils import show_exception

from .link import Link
from .session import ClientsSession

logger = logger.bind(scheme="telenotifier")


class TelegramStream(io.TextIOWrapper):
    """消息推送处理器类"""

    def __init__(self, account: TelegramAccount, instant=False):
        super().__init__(io.BytesIO(), line_buffering=True)
        self.account = account
        self.instant = instant

        self.queue = asyncio.Queue()
        self.watch = asyncio.create_task(self.watchdog())

    async def watchdog(self):
        while True:
            message = await self.queue.get()
            try:
                result = await asyncio.wait_for(self.send(message), 20)
            except asyncio.TimeoutError:
                logger.warning("推送消息到 Telegram 超时.")
            except Exception as e:
                logger.warning("推送消息到 Telegram 失败.")
                show_exception(e)
            else:
                if not result:
                    logger.warning(f"推送消息到 Telegram 失败.")
            finally:
                self.queue.task_done()

    async def send(self, message):
        async with ClientsSession([self.account]) as clients:
            async for _, tg in clients:
                if self.instant:
                    return await Link(tg).send_msg(message)
                else:
                    return await Link(tg).send_log(message)
            else:
                return False

    def write(self, message):
        message = Text.from_markup(message).plain
        if message.endswith("\n"):
            message = message[:-1]
        if message:
            self.queue.put_nowait(message)

    async def join(self):
        await self.queue.join()
        self.watch.cancel()



import random
import re
import string
from typing import List, Union
from loguru import logger
from pydantic import BaseModel, ValidationError
from pyrogram.types import Message

from . import Monitor
from ..embyboss import EmbybossRegister

__ignore__ = True


class TemplateAMonitorConfig(BaseModel):
    name: str = None
    chat_name: Union[str, int] = None  # 监控的群聊名称
    chat_allow_outgoing: bool = False  # 是否支持自己发言触发
    chat_user: Union[str, List[str]] = []  # 仅被列表中用户的发言触发 (支持 username / userid)
    chat_keyword: Union[str, List[str]] = []  # 仅当消息含有列表中的关键词时触发, 支持 regex
    chat_except_keyword: Union[str, List[str]] = []  # 消息含有列表中的关键词时不触发, 支持 regex
    chat_probability: float = 1.0  # 发信概率 (0最低, 1最高)
    chat_delay: int = 0  # 发信延迟 (s)
    chat_follow_user: int = 0  # 需要等待 N 个用户发送 {chat_reply} 方可回复
    chat_reply: str = None  # 回复的内容, 可以为恒定字符串或函数或异步函数
    allow_edit: bool = False  # 编辑消息内容后也触发
    trigger_interval: float = 2  # 每次触发的最低时间间隔
    trigger_sim: int = 1  # 同时触发的最大并行数
    trigger_max_time: float = 120  # 触发后处理的最长时间
    allow_caption: bool = True  # 是否允许带照片的消息
    allow_text: bool = True  # 是否允许不带照片的消息
    send: bool = True  # 是否发送通知
    send_immediately: bool = True  # 是否发送即时日志, 不等待每日推送时间
    try_register_bot: str = None  # 尝试注册的机器人名称 (需为: https://github.com/berry8838/Sakura_embyboss)


class TemplateAMonitor(Monitor):
    init_first = True
    additional_auth = ["prime"]
    notify_create_name = True
    debug_no_log = True

    async def init(self):
        try:
            self.t_config = TemplateAMonitorConfig.model_validate(self.config)
        except ValidationError as e:
            self.log.warning(f"初始化失败: 监控器自定义模板 A 的配置错误:\n{e}")
            return False
        self.name = self.t_config.name or "自定义"
        self.chat_name = self.t_config.chat_name
        self.chat_allow_outgoing = self.t_config.chat_allow_outgoing
        self.chat_user = self.t_config.chat_user
        self.chat_keyword = self.t_config.chat_keyword
        self.chat_except_keyword = self.t_config.chat_except_keyword
        self.chat_probability = self.t_config.chat_probability
        self.chat_delay = self.t_config.chat_delay
        self.chat_follow_user = self.t_config.chat_follow_user
        self.chat_reply = self.t_config.chat_reply
        self.allow_edit = self.t_config.allow_edit
        self.trigger_interval = self.t_config.trigger_interval
        self.trigger_sim = self.t_config.trigger_sim
        self.trigger_max_time = self.t_config.trigger_max_time
        self.allow_caption = self.t_config.allow_caption
        self.allow_text = self.t_config.allow_text
        if (not self.chat_keyword) and (not self.chat_user) and (not self.chat_name):
            self.log.warning(f"初始化失败: 没有定义任何监控项, 请参考教程进行配置.")
            return False
        self.log = logger.bind(scheme="telemonitor", name=self.name, username=self.client.me.name)
        return True

    async def on_trigger(self, message: Message, key, reply):
        content = message.text or message.caption
        if self.t_config.send:
            if message.from_user:
                msg = f"监控器 {self.name} 收到来自 {message.from_user.name} 的关键消息: {content}"
            else:
                msg = f"监控器 {self.name} 收到关键消息: {content}"
            if self.t_config.send_immediately:
                self.log.bind(msg=True).info(msg)
            else:
                self.log.bind(log=True).info(msg)
        if self.t_config.try_register_bot:
            random_code = "".join(random.choices(string.ascii_letters + string.digits, k=4))
            if await EmbybossRegister(self.client, self.log, self.unique_name, random_code).run(
                self.t_config.try_register_bot
            ):
                self.log.bind(log=True).info(
                    f"监控器 {self.name} 成功注册机器人 {self.t_config.try_register_bot}."
                )
        else:
            await super().on_trigger(message, key, reply)

    def get_unique_name(self):
        if not self.t_config.try_register_bot:
            return None
        unique_name = self.config.get("unique_name", None)
        if unique_name:
            self.log.info(f'根据您的设置, 当监控到开注时, 该站点将以用户名 "{unique_name}" 注册.')
            if not re.search("^\w+$", unique_name):
                self.log.warning(f"用户名含有除 a-z, A-Z, 0-9, 以及下划线之外的字符, 可能导致注册失败.")
            return unique_name
        else:
            return Monitor.unique_cache[self.client.me]


def use(**kw):
    return type("TemplatedClass", (TemplateAMonitor,), kw)



from . import BotCheckin

__ignore__ = True


class ByteVirtGroupCheckin(BotCheckin):
    name = "ByteVirt 群组发言"
    chat_name = "bytevirtchat"
    bot_checkin_cmd = ["/showmethemoney"]
    bot_account_fail_keywords = ["未绑定"]



import asyncio
from io import BytesIO
import random
import re

from curl_cffi.requests import AsyncSession, Response
from pyrogram.types import Message
from pyrogram.errors import MessageIdInvalid
from PIL import Image
import numpy as np

from embykeeper.config import config
from embykeeper.utils import show_exception, get_proxy_str

from ..lock import pornemby_alert
from . import Monitor

JAVDATABASE_URL = "https://www.javdatabase.com"


class _PornembyExamResultMonitor(Monitor):
    name = "Pornemby 科举答案"
    chat_keyword = r"问题\d*：(.*?)\n+答案为：([ABCD])\n+([A-Z-\d]+)"
    additional_auth = ["pornemby_pack"]
    allow_edit = True

    async def on_trigger(self, message: Message, key, reply):
        self.log.info(f"本题正确答案为 {key[1]} ({key[2]}).")


class _PornembyExamAnswerMonitor(Monitor):
    name = "Pornemby 科举"
    chat_user = [
        "pornemby_question_bot",
        "PronembyTGBot2_bot",
        "PronembyTGBot3_bot",
        "PornembyBot",
        "Porn_Emby_Bot",
        "Porn_Emby_Scriptbot",
    ]
    chat_keyword = (
        r"问题\d*：根据以上封面图，猜猜是什么番号？\n+A:(.*)\n+B:(.*)\n+C:(.*)\n+D:(.*)\n(?!\n*答案)"
    )
    additional_auth = ["pornemby_pack"]
    allow_edit = True

    key_map = {
        "A": ["A", "🅰"],
        "B": ["B", "🅱"],
        "C": ["C", "🅲"],
        "D": ["D", "🅳"],
    }

    async def use_cfsolver(self):
        from embykeeper.cloudflare import get_cf_clearance

        if self.proxy:
            if self.proxy.scheme != "socks5":
                self.log.warning(f"站点验证解析仅支持 SOCKS5 代理，由于当前代理协议不支持, 将尝试不使用代理.")
                self.proxy = None
            else:
                self.log.info(
                    f"验证码解析将使用代理, 可能导致解析失败, 若失败请使用"
                    '"use_proxy = false" 以禁用该站点的代理.'
                )
        try:
            cf_clearance, useragent = await get_cf_clearance(JAVDATABASE_URL, self.proxy)
            if not cf_clearance:
                self.log.warning(f"Cloudflare 验证码解析失败.")
                return False
            else:
                self.cf_clearance = cf_clearance
                self.useragent = useragent
                return True
        except Exception as e:
            self.log.warning(f"Cloudflare 验证码解析时出现错误.")
            show_exception(e, regular=False)
            return False

    async def init(self):
        self.proxy = config.proxy
        self.useragent = None
        self.cf_clearance = None
        for _ in range(3):
            try:
                if self.useragent:
                    headers = {"User-Agent": self.useragent}
                else:
                    headers = None
                if self.cf_clearance:
                    cookies = {"cf_clearance": self.cf_clearance}
                else:
                    cookies = None
                async with AsyncSession(
                    proxy=get_proxy_str(self.proxy),
                    impersonate="chrome",
                    timeout=10.0,
                    allow_redirects=True,
                    headers=headers,
                    cookies=cookies,
                ) as session:
                    resp: Response = await session.get(JAVDATABASE_URL)
                    if resp.status_code == 403 and (
                        "cf-wrapper" in resp.text or "Just a moment" in resp.text
                    ):
                        if self.cf_clearance:
                            self.log.warning("初始化失败: Javdatabase 在 Cloudflare 验证码解析后依然有验证")
                            return False
                        self.log.info("Javdatabase 存在 Cloudflare 保护，正在尝试解析.")
                        await self.use_cfsolver()
                        continue
                    elif not resp.ok:
                        self.log.warning(f"初始化失败: Javdatabase 返回状态码错误: {resp.status_code}.")
                        return False
                    return True
            except Exception as e:
                self.log.warning(f"初始化失败: 无法连接 Javdatabase: {e.__class__.__name__}: {str(e)}")
                return False

    async def get_cover_image_javdatabase(self, code: str):
        # 添加重试次数
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:
                async with AsyncSession(
                    proxy=get_proxy_str(self.proxy),
                    impersonate="chrome",
                    timeout=10.0,
                    allow_redirects=True,
                ) as session:
                    detail_url = f"{JAVDATABASE_URL}/movies/{code.lower()}/"
                    resp: Response = await session.get(detail_url)
                    if resp.status_code != 200:
                        self.log.warning(
                            f"获取影片详情失败: 网址访问错误: {detail_url} ({resp.status_code})."
                        )
                        retry_count += 1
                        if retry_count < max_retries:
                            self.log.info(f"正在进行第 {retry_count + 1} 次重试...")
                            continue
                        return None
                    html = resp.text
                    pattern = f'<div id="thumbnailContainer".*({JAVDATABASE_URL}//covers/thumb/.*/.*.webp)'
                    match = re.search(pattern, html)
                    if not match:
                        self.log.warning(f"获取封面图片失败: 未找到图片: {detail_url} ({resp.status_code}).")
                        return None
                    img_url = match.group(1)
                    # 下载封面图片
                    img_response = await session.get(img_url)
                    if img_response.status_code == 200:
                        return BytesIO(img_response.content)
                    else:
                        self.log.warning(
                            f"获取封面图片失败: 网址访问错误: {img_url} ({img_response.status_code})."
                        )
                        return None

            except Exception as e:
                retry_count += 1
                if retry_count < max_retries:
                    self.log.info(
                        f"获取封面图片失败，正在进行第 {retry_count + 1} 次重试: {e.__class__.__name__}: {str(e)}"
                    )
                    continue
                self.log.warning(f"获取封面图片失败: {e.__class__.__name__}: {str(e)}")
                show_exception(e)
                return None

            # 如果执行到这里说明成功获取了图片，直接返回
            break

        return None

    async def get_cover_image_r18_dev(self, code: str):
        max_retries = 3
        retry_count = 0

        while retry_count < max_retries:
            try:
                async with AsyncSession(
                    proxy=get_proxy_str(self.proxy),
                    timeout=10.0,
                    allow_redirects=True,
                    impersonate="chrome110",
                ) as session:
                    # 先获取 content_id
                    detail_url = f"https://r18.dev/videos/vod/movies/detail/-/dvd_id={code.lower()}/json"
                    # 获取 content_id
                    resp: Response = await session.get(detail_url)
                    if resp.status_code != 200:
                        self.log.warning(
                            f"获取影片详情失败: 网址访问错误: {detail_url} ({resp.status_code})."
                        )
                        retry_count += 1
                        if retry_count < max_retries:
                            self.log.info(f"正在进行第 {retry_count + 1} 次重试...")
                            continue
                        return None
                    detail_json = resp.json()
                    content_id = detail_json.get("content_id")
                    if not content_id:
                        self.log.warning(f"获取影片详情失败: 无法获取 content_id: {detail_url}")
                        return None

                    # 获取封面图片 URL
                    combined_url = f"https://r18.dev/videos/vod/movies/detail/-/combined={content_id}/json"
                    resp: Response = await session.get(combined_url)
                    if resp.status_code != 200:
                        self.log.warning(
                            f"获取封面详情失败: 网址访问错误: {combined_url} ({resp.status_code})."
                        )
                        return None
                    combined_json = resp.json()
                    jacket_url = combined_json.get("jacket_thumb_url")
                    if not jacket_url:
                        self.log.warning(f"获取封面详情失败: 无法获取封面URL: {combined_url}")
                        return None

                    # 下载封面图片
                    img_response = await session.get(jacket_url)
                    if img_response.status_code == 200:
                        return BytesIO(img_response.content)
                    else:
                        self.log.warning(
                            f"获取封面图片失败: 网址访问错误: {jacket_url} ({img_response.status_code})."
                        )
                        return None

            except Exception as e:
                retry_count += 1
                if retry_count < max_retries:
                    self.log.info(
                        f"获取封面图片失败，正在进行第 {retry_count + 1} 次重试: {e.__class__.__name__}: {str(e)}"
                    )
                    continue
                self.log.warning(f"获取封面图片失败: {e.__class__.__name__}: {str(e)}")
                show_exception(e)
                return None

        return None

    def compare_images(self, img1_bytes: BytesIO, img2_bytes: BytesIO) -> float:
        try:
            img1 = Image.open(img1_bytes).convert("RGB").resize((100, 100))
            img2 = Image.open(img2_bytes).convert("RGB").resize((100, 100))

            arr1 = np.array(img1)
            arr2 = np.array(img2)
            mse = np.mean((arr1 - arr2) ** 2)

            similarity = 1 / (1 + mse)
            return similarity
        except Exception as e:
            self.log.debug(f"图片比较失败: {e}")
            return 0

    async def on_trigger(self, message: Message, key, reply):
        if not message.photo or not message.reply_markup:
            return
        if pornemby_alert.get(self.client.me.id, False):
            self.log.info(f"由于风险急停不作答.")
            return
        if random.random() > self.config.get("possibility", 1.0):
            self.log.info(f"由于概率设置不作答.")
            return

        question_photo = await message.download(in_memory=True)

        codes = [re.sub(r"-\w$", "", k) for k in key]

        async def get_cover_with_timeout(code):
            try:
                return code, await asyncio.wait_for(self.get_cover_image_javdatabase(code), timeout=10)
            except asyncio.TimeoutError:
                self.log.debug(f"获取 {code} 封面超时")
                return code, None

        cover_tasks = [get_cover_with_timeout(code) for code in codes]
        covers = await asyncio.gather(*cover_tasks)
        max_similarity = -1
        best_code = None
        for code, cover in covers:
            if cover is None:
                continue
            question_photo.seek(0)
            cover.seek(0)
            similarity = self.compare_images(question_photo, cover)
            self.log.debug(f"番号 {code} 相似度: {similarity:.4f}")
            if similarity > max_similarity:
                max_similarity = similarity
                best_code = code
        if best_code:
            result = ["A", "B", "C", "D"][codes.index(best_code)]
            self.log.info(f"选择相似度最高的番号: {best_code} ({result}) (相似度: {max_similarity:.4f})")
            buttons = [k.text for r in message.reply_markup.inline_keyboard for k in r]
            answer_options = self.key_map[result]
            for button_text in buttons:
                if any((o in button_text) for o in answer_options):
                    try:
                        await message.click(button_text)
                    except (TimeoutError, MessageIdInvalid):
                        pass
                    break
            else:
                self.log.info(f"点击失败: 未找到匹配的按钮文本 {result}.")
        else:
            self.log.warning("未找到匹配的封面图片")


class PornembyExamMonitor:
    class PornembyExamResultMonitor(_PornembyExamResultMonitor):
        chat_name = "Pornemby"

    class PornembyExamAnswerMonitor(_PornembyExamAnswerMonitor):
        chat_name = "Pornemby"
